import inspect
import threading
import time
import traceback
from concurrent.futures import ThreadPoolExecutor

import ophyd
from ophyd import Kind, OphydObject, Staged
from ophyd.utils import errors as ophyd_errors

from bec_lib import Alarms, BECService, MessageEndpoints, bec_logger, messages
from bec_lib.connector import ConnectorBase
from bec_lib.device import OnFailure
from bec_lib.messages import BECStatus
from bec_server.device_server.devices.devicemanager import DeviceManagerDS
from bec_server.device_server.rpc_mixin import RPCMixin

logger = bec_logger.logger

register_stop = threading.Event()


class DisabledDeviceError(Exception):
    """Exception raised when a disabled device is accessed"""


class InvalidDeviceError(Exception):
    """Exception raised when an invalid device is accessed"""


class DeviceServer(RPCMixin, BECService):
    """DeviceServer using ophyd as a service
    This class is intended to provide a thin wrapper around ophyd and the devicemanager. It acts as the entry point for other services
    """

    def __init__(self, config, connector_cls: ConnectorBase) -> None:
        super().__init__(config, connector_cls, unique_service=True)
        self._tasks = []
        self.device_manager = None
        self.connector.register(
            MessageEndpoints.scan_queue_modification(), cb=self.register_interception_callback
        )
        self.executor = ThreadPoolExecutor(max_workers=4)
        self._start_device_manager()

    def _start_device_manager(self):
        self.device_manager = DeviceManagerDS(self, status_cb=self.update_status)
        self.device_manager.initialize(self.bootstrap_server)

    def start(self) -> None:
        """start the device server"""
        if register_stop.is_set():
            register_stop.clear()

        self.connector.register(
            MessageEndpoints.device_instructions(),
            event=register_stop,
            cb=self.instructions_callback,
            parent=self,
        )

        self.status = BECStatus.RUNNING

    def update_status(self, status: BECStatus):
        """update the status of the device server"""
        self.status = status

    def stop(self) -> None:
        """stop the device server"""
        register_stop.set()
        self.status = BECStatus.IDLE

    def shutdown(self) -> None:
        """shutdown the device server"""
        super().shutdown()
        self.stop()
        self.device_manager.shutdown()

    def _update_device_metadata(self, instr) -> None:
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]
        for dev in devices:
            device_root = dev.split(".")[0]
            self.device_manager.devices.get(device_root).metadata = instr.metadata

    def register_interception_callback(self, msg, **_kwargs) -> None:
        """callback for receiving scan modifications / interceptions"""
        mvalue = msg.value
        if mvalue is None:
            logger.warning("Failed to parse scan queue modification message.")
            return
        logger.info(f"Receiving: {mvalue.content}")
        if mvalue.content.get("action") in ["pause", "abort", "halt"]:
            self.stop_devices()

    def stop_devices(self) -> None:
        """stop all enabled devices"""
        logger.info("Stopping devices after receiving 'abort' request.")
        self.status = BECStatus.BUSY
        for dev in self.device_manager.devices.enabled_devices:
            if dev.read_only:
                # don't stop devices that we haven't set
                continue
            if hasattr(dev.obj, "stop"):
                dev.obj.stop()
        self.status = BECStatus.RUNNING

    def _assert_device_is_enabled(self, instructions: messages.DeviceInstructionMessage) -> None:
        devices = instructions.content["device"]

        if isinstance(devices, str):
            devices = [devices]

        for dev in devices:
            dev = dev.split(".")[0]
            if not self.device_manager.devices[dev].enabled:
                raise DisabledDeviceError(f"Cannot access disabled device {dev}.")

    def _assert_device_is_valid(self, instructions: messages.DeviceInstructionMessage) -> None:
        devices = instructions.content["device"]
        if not devices:
            raise InvalidDeviceError("At least one device must be specified.")
        if isinstance(devices, str):
            devices = [devices]
        for dev in devices:
            dev = dev.split(".")[0]
            if dev not in self.device_manager.devices:
                raise InvalidDeviceError(f"There is no device with the name {dev}.")

    def handle_device_instructions(self, msg: str) -> None:
        """Parse a device instruction message and handle the requested action. Action
        types are set, read, rpc, kickoff or trigger.

        Args:
            msg (str): A DeviceInstructionMessage string containing the action and its parameters

        """
        action = None
        try:
            instructions = msg
            if not instructions.content["device"]:
                return
            action = instructions.content["action"]
            self._assert_device_is_valid(instructions)
            if action != "rpc":
                # rpc has its own error handling
                self._assert_device_is_enabled(instructions)
            self._update_device_metadata(instructions)

            if action == "set":
                self._set_device(instructions)
            elif action == "read":
                self._read_device(instructions)
            elif action == "rpc":
                self.run_rpc(instructions)
            elif action == "kickoff":
                self._kickoff_device(instructions)
            elif action == "complete":
                self._complete_device(instructions)
            elif action == "trigger":
                self._trigger_device(instructions)
            elif action == "stage":
                self._stage_device(instructions)
            elif action == "unstage":
                self._unstage_device(instructions)
            elif action == "pre_scan":
                self._pre_scan(instructions)
            else:
                logger.warning(f"Received unknown device instruction: {instructions}")
        except ophyd_errors.LimitError as limit_error:
            content = traceback.format_exc()
            logger.error(content)
            self.connector.raise_alarm(
                severity=Alarms.MAJOR,
                source=instructions.content,
                msg=content,
                alarm_type=limit_error.__class__.__name__,
                metadata=instructions.metadata,
            )
        except Exception as exc:  # pylint: disable=broad-except
            if action == "rpc":
                self._send_rpc_exception(exc, instructions)
            else:
                content = traceback.format_exc()
                self.connector.log_error({"source": msg, "message": content})
                logger.error(content)
                self.connector.raise_alarm(
                    severity=Alarms.MAJOR,
                    source=instructions.content,
                    msg=content,
                    alarm_type=exc.__class__.__name__,
                    metadata=instructions.metadata,
                )

    @staticmethod
    def instructions_callback(msg, *, parent, **_kwargs) -> None:
        """callback for handling device instructions"""
        parent.executor.submit(parent.handle_device_instructions, msg.value)

    def _trigger_device(self, instr: messages.DeviceInstructionMessage) -> None:
        logger.debug(f"Trigger device: {instr}")
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]
        for dev in devices:
            obj = self.device_manager.devices.get(dev)
            obj.metadata = instr.metadata
            status = obj.obj.trigger()
            status.__dict__["instruction"] = instr
            status.add_callback(self._status_callback)

    def _kickoff_device(self, instr: messages.DeviceInstructionMessage) -> None:
        logger.debug(f"Kickoff device: {instr}")
        obj = self.device_manager.devices.get(instr.content["device"]).obj
        kickoff_args = inspect.getfullargspec(obj.kickoff).args
        kickoff_parameter = instr.content["parameter"].get("configure", {})
        if len(kickoff_args) > 1:
            obj.kickoff(metadata=instr.metadata, **kickoff_parameter)
            return
        obj.configure(kickoff_parameter)
        status = obj.kickoff()
        status.__dict__["instruction"] = instr
        status.add_callback(self._status_callback)

    def _complete_device(self, instr: messages.DeviceInstructionMessage) -> None:
        if instr.content["device"] is None:
            devices = [dev.name for dev in self.device_manager.devices.enabled_devices]
        else:
            devices = instr.content["device"]
            if not isinstance(devices, list):
                devices = [devices]
        for dev in devices:
            obj = self.device_manager.devices.get(dev).obj
            if not hasattr(obj, "complete"):
                # if the device does not have a complete method, we assume that it is done
                status = ophyd.StatusBase()
                status.obj = obj
                status.set_finished()
            else:
                logger.debug(f"Completing device: {dev}")
                status = obj.complete()
            if status is None:
                raise InvalidDeviceError(
                    f"The complete method of device {dev} does not return a DeviceStatus object."
                )
            status.__dict__["instruction"] = instr
            status.add_callback(self._status_callback)

    def _set_device(self, instr: messages.DeviceInstructionMessage) -> None:
        device_obj = self.device_manager.devices.get(instr.content["device"])
        if device_obj.read_only:
            raise DisabledDeviceError(
                f"Setting the device {device_obj.name} is currently disabled."
            )
        logger.debug(f"Setting device: {instr}")
        val = instr.content["parameter"]["value"]
        obj = self.device_manager.devices.get(instr.content["device"]).obj
        # self.device_manager.add_req_done_sub(obj)
        status = obj.set(val)
        status.__dict__["instruction"] = instr
        status.add_callback(self._status_callback)

    def _pre_scan(self, instr: messages.DeviceInstructionMessage) -> None:
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]
        pipe = self.connector.pipeline()
        for dev in devices:
            obj = self.device_manager.devices.get(dev)
            obj.metadata = instr.metadata
            if hasattr(obj.obj, "pre_scan"):
                obj.obj.pre_scan()
            dev_msg = messages.DeviceReqStatusMessage(
                device=dev, success=True, metadata=instr.metadata
            )
            self.connector.set(MessageEndpoints.device_req_status(dev), dev_msg, pipe)
        pipe.execute()

    def _status_callback(self, status):
        pipe = self.connector.pipeline()
        if hasattr(status, "device"):
            obj = status.device
        else:
            obj = status.obj
        device_name = obj.root.name
        dev_msg = messages.DeviceReqStatusMessage(
            device=device_name, success=status.success, metadata=status.instruction.metadata
        )
        logger.debug(f"req status for device {device_name}: {status.success}")
        self.connector.set(MessageEndpoints.device_req_status(device_name), dev_msg, pipe)
        response = status.instruction.metadata.get("response")
        if response:
            self.connector.lpush(
                MessageEndpoints.device_req_status_container(status.instruction.metadata["RID"]),
                dev_msg,
                pipe,
                expire=18000,
            )
        content = status.instruction.content
        is_config_set = content["action"] == "set"
        is_rpc_set = content["action"] == "rpc" and (".set" in content["parameter"]["func"])

        if is_config_set or is_rpc_set:
            if obj.kind == Kind.config:
                self._update_read_configuration(obj, status.instruction.metadata, pipe)
            elif obj.kind in [Kind.normal, Kind.hinted]:
                self._read_device(status.instruction)
        pipe.execute()

    def _update_read_configuration(self, obj: OphydObject, metadata: dict, pipe) -> None:
        dev_config_msg = messages.DeviceMessage(
            signals=obj.root.read_configuration(), metadata=metadata
        )
        self.connector.set_and_publish(
            MessageEndpoints.device_read_configuration(obj.root.name), dev_config_msg, pipe
        )

    def _read_device(self, instr: messages.DeviceInstructionMessage) -> None:
        # check performance -- we might have to change it to a background thread
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]

        self._read_and_update_devices(devices, instr.metadata)

    def _read_and_update_devices(self, devices: list[str], metadata: dict) -> list:
        start = time.time()
        pipe = self.connector.pipeline()
        signal_container = []
        for dev in devices:
            device_root = dev.split(".")[0]
            self.device_manager.devices.get(device_root).metadata = metadata
            obj = self.device_manager.devices.get(device_root).obj
            try:
                signals = obj.read()
                signal_container.append(signals)
            except Exception as exc:
                signals = self._retry_obj_method(dev, obj, "read", exc)

            self.connector.set_and_publish(
                MessageEndpoints.device_read(device_root),
                messages.DeviceMessage(signals=signals, metadata=metadata),
                pipe,
            )
            self.connector.set_and_publish(
                MessageEndpoints.device_readback(device_root),
                messages.DeviceMessage(signals=signals, metadata=metadata),
                pipe,
            )
            self.connector.set(
                MessageEndpoints.device_status(device_root),
                messages.DeviceStatusMessage(device=device_root, status=0, metadata=metadata),
                pipe,
            )
        pipe.execute()
        logger.debug(
            f"Elapsed time for reading and updating status info: {(time.time()-start)*1000} ms"
        )
        return signal_container

    def _read_config_and_update_devices(self, devices: list[str], metadata: dict) -> list:
        start = time.time()
        pipe = self.connector.pipeline()
        signal_container = []
        for dev in devices:
            self.device_manager.devices.get(dev).metadata = metadata
            obj = self.device_manager.devices.get(dev).obj
            try:
                signals = obj.read_configuration()
                signal_container.append(signals)
            except Exception as exc:
                signals = self._retry_obj_method(dev, obj, "read_configuration", exc)
            self.connector.set_and_publish(
                MessageEndpoints.device_read_configuration(dev),
                messages.DeviceMessage(signals=signals, metadata=metadata),
                pipe,
            )
        pipe.execute()
        logger.debug(
            f"Elapsed time for reading and updating status info: {(time.time()-start)*1000} ms"
        )
        return signal_container

    def _retry_obj_method(self, device: str, obj: OphydObject, method: str, exc: Exception) -> dict:
        self.device_manager.connector.raise_alarm(
            severity=Alarms.WARNING,
            alarm_type="Warning",
            source="DeviceServer",
            msg=f"Failed to run {method} on device {device}.",
            metadata={},
        )
        device_root = device.split(".")[0]
        ds_dev = self.device_manager.devices.get(device_root)
        if ds_dev.on_failure == OnFailure.RAISE:
            raise exc

        if ds_dev.on_failure == OnFailure.RETRY:
            # try to read it again, may have been only a glitch
            signals = getattr(obj, method)()
        elif ds_dev.on_failure == OnFailure.BUFFER:
            # if possible, fall back to past readings
            logger.warning(
                f"Failed to run {method} on device {device_root}. Trying to load an old value."
            )
            if method == "read":
                old_msg = self.connector.get(MessageEndpoints.device_read(device_root))
            elif method == "read_configuration":
                old_msg = self.connector.get(
                    MessageEndpoints.device_read_configuration(device_root)
                )
            else:
                raise ValueError(f"Unknown method {method}.")
            if not old_msg:
                raise exc
            signals = old_msg.content["signals"]
        return signals

    def _stage_device(self, instr: messages.DeviceInstructionMessage) -> None:
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]

        pipe = self.connector.pipeline()
        for dev in devices:
            obj = self.device_manager.devices[dev].obj
            if hasattr(obj, "_staged"):
                # pylint: disable=protected-access
                if obj._staged == Staged.yes:
                    logger.info(f"Device {obj.name} was already staged and will be first unstaged.")
                    self.device_manager.devices[dev].obj.unstage()
                self.device_manager.devices[dev].obj.stage()
            self.connector.set(
                MessageEndpoints.device_staged(dev),
                messages.DeviceStatusMessage(device=dev, status=1, metadata=instr.metadata),
                pipe,
            )
        pipe.execute()

    def _unstage_device(self, instr: messages.DeviceInstructionMessage) -> None:
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]

        pipe = self.connector.pipeline()
        for dev in devices:
            obj = self.device_manager.devices[dev].obj
            if hasattr(obj, "_staged"):
                # pylint: disable=protected-access
                if obj._staged == Staged.yes:
                    self.device_manager.devices[dev].obj.unstage()
                else:
                    logger.debug(f"Device {obj.name} was already unstaged.")
            self.connector.set(
                MessageEndpoints.device_staged(dev),
                messages.DeviceStatusMessage(device=dev, status=0, metadata=instr.metadata),
                pipe,
            )
        pipe.execute()



""" Module for file utilities."""

from __future__ import annotations

import os
import warnings
from typing import TYPE_CHECKING

from bec_lib.bec_errors import ServiceConfigError
from bec_lib.endpoints import MessageEndpoints
from bec_lib.messages import ScanStatusMessage

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector


class ServiceConfigParser:
    """Service Config Parser"""

    def __init__(self, service_config: dict = None) -> None:
        """Initialize the service config parser.

        Args:
            service_config (dict): Service config from BEC startup.
        """
        self._service_config = service_config

    @property
    def service_config(self) -> dict:
        """Get the service config."""
        return self._service_config

    def get_base_path(self) -> str:
        """Get the base path from the config."""
        if not self.service_config:
            raise ServiceConfigError("Service config must contain a file writer definition.")
        if not self.service_config.get("base_path"):
            raise ServiceConfigError("File writer config must define a base path.")
        return os.path.abspath(os.path.expanduser(self.service_config.get("base_path")))

    def create_directory(self, directory: str) -> None:
        """Create a directory if it does not exist."""
        os.makedirs(directory, exist_ok=True)


class LogWriter:
    """Log writer class"""

    def __init__(self, service_config: dict) -> None:
        """
        Initialize the log writer class.

        Args:
            service_config (dict): Log writer service config. Must at least contain a base_path.
        """
        self.service_config_parser = ServiceConfigParser(service_config)
        self._base_path = self.service_config_parser.get_base_path()
        self._directory = os.path.join(self._base_path, "logs")
        self.create_directory(self.directory)

    def create_directory(self, fname: str = None) -> None:
        """Create the log directory."""
        self.service_config_parser.create_directory(fname)

    @property
    def directory(self) -> str:
        """Get the log directory.

        Returns:
            str: String representation of log directory
        """
        return self._directory


class DeviceConfigWriter:
    """Device config writer class"""

    def __init__(self, service_config: dict) -> None:
        """
        Initialize the device config writer class.
        At the moment, this uses the same base_path as the log writer

        Args:
            service_config (dict): Device config writer service config. Must at least contain a base_path.
        """
        self.service_config_parser = ServiceConfigParser(service_config)
        self._base_path = self.service_config_parser.get_base_path()
        self._directory = os.path.join(self._base_path, "device_configs")
        self.create_directory(fname=self.directory)

    def create_directory(self, fname: str = None) -> None:
        """Create the device config directory."""
        self.service_config_parser.create_directory(fname)

    @property
    def directory(self) -> str:
        """Get the device config directory.

        Returns:
            str: String representation of device config directory
        """
        return self._directory

    def get_recovery_directory(self) -> str:
        """
        Compile the recovery config directory.
        """
        return os.path.join(self.directory, "recovery_configs")


class FileWriterError(Exception):
    """Exception for errors in the file writer"""


class FileWriter:
    """FileWriter for creating file paths and directories for services and devices."""

    def __init__(
        self,
        *args,
        service_config: dict = None,
        connector: RedisConnector = None,
        scan_bundle: int = 1000,
        leading_zeros: int = 5,
        **kwargs,
    ) -> None:
        """
        Initialize the file writer mixin class.
        If no RedisConnector is provided, the class will not be able to communicate with the REDIS server.

        In this case, it will fall back to a default base path in the current directory.

        Args:
            service_config (dict): File writer service config. Must at least contain base_path.
            connector (Redisconnector, optional): Connector class to use. Defaults to RedisConnector.
            scan_bundle (int, optional): Scan bundle size. Defaults to 1000.
            leading_zeros (int, optional): Number of leading zeros. Defaults to 5.

        """
        self.service_config = service_config
        self._scan_bundle = scan_bundle
        self._leading_zeros = leading_zeros
        self._kwargs = kwargs
        self._base_path = "."  # default to current directory
        self.connector = connector
        self.service_config_parser = None
        self._configured = False
        self._initialized = True
        if self.connector:
            self.configure_file_writer()
            self._configured = True

    def configure_file_writer(self):
        """Configure the file writer mixin class in case service_config is provided"""
        self.service_config_parser = ServiceConfigParser(self.service_config)
        self._base_path = self.service_config_parser.get_base_path()

    @property
    def scan_bundle(self) -> int:
        """Get the scan bundle size."""
        return self._scan_bundle

    @property
    def leading_zeros(self) -> int:
        """Get the number of leading zeros."""
        return self._leading_zeros

    def get_scan_directory(
        self, scan_number: int, scan_bundle: int, leading_zeros: int = None, user_suffix: str = None
    ) -> str:
        """
        Get the scan directory for a given scan number and scan bundle.

        Args:
            scan_bundle (int): Scan bundle size
            scan_number (int): Scan number
            leading_zeros (int, optional): Number of leading zeros. Defaults to None.
            user_suffix (str, optional): User defined suffix. Defaults to None.

        Returns:
            str: Scan directory

        Examples:
            >>> get_scan_directory(1234, 1000, 5)
            'S01000-01999/S01234'
            >>> get_scan_directory(1234, 1000, 5, 'sampleA')
            'S01000-01999/S01234_sampleA'
        """
        if leading_zeros is None:
            leading_zeros = len(str(scan_bundle))
        floor_dir = scan_number // scan_bundle * scan_bundle
        rtr = f"S{floor_dir:0{leading_zeros}d}-{floor_dir+scan_bundle-1:0{leading_zeros}d}/S{scan_number:0{leading_zeros}d}"
        if user_suffix:
            rtr += f"_{user_suffix}"
        return rtr

    def get_scan_msg(self):
        """Get the scan message for the next scan"""
        msg = self.connector.get(MessageEndpoints.scan_status())
        if not isinstance(msg, ScanStatusMessage):
            return None
        return msg

    def compile_full_filename(self, suffix: str, create_dir: bool = True) -> str:
        """
        Compile a full filename for a given scan number and suffix.

        This method should only be called after a scan has been opened,
        i.e. preferable in stage and not during __init__.

        The method will use the last scan message received in REDIS,
        and will return an empty string if no scan message is available.

        Args:
            suffix (str): Filename suffix including extension. We allow alphanumeric, ascii characters - and _.
            file_type (str) : Optional, will default to h5.
            create_dir (bool, optional): Create the directory if it does not exist. Defaults to True.

        Returns:
            str: Full filename
        """
        # to check if suffix is alphanumeric and ascii, however we allow in addition - and _
        check_suffix = suffix.replace("_", "").replace("-", "")
        if not check_suffix.isalnum() or not check_suffix.isascii():
            raise FileWriterError(
                f"Can't use suffix {suffix}; formatting is alphanumeric:{suffix.isalnum()} and ascii {suffix.isascii()}"
            )
        if self._configured:
            scan_msg = self.get_scan_msg()
            if not scan_msg:
                warnings.warn("No scan message available.")
                return ""
            scannr = scan_msg.content["info"]["scan_number"]
            user_suffix = scan_msg.info["file_writer_data"].get("file_suffix")
            scan_dir = scan_msg.info["file_writer_data"].get("file_directory")
            if not scan_dir:
                scan_dir = self.get_scan_directory(
                    scannr, self.scan_bundle, self.leading_zeros, user_suffix=user_suffix
                )
            if user_suffix:
                suffix += f"_{user_suffix}"
            full_file = os.path.join(
                self._base_path, "data", scan_dir, f"S{scannr:0{self._leading_zeros}d}_{suffix}.h5"
            )
        else:
            full_file = os.path.join(self._base_path, "data", f"S00000_default_{suffix}.h5")
            warnings.warn(f"No service config provided, using default base path {full_file}.")
        if create_dir:
            os.makedirs(os.path.dirname(full_file), exist_ok=True)
        return full_file



"""
This module provides the BeamlineChecks class, which is used to perform beamline checks. It also provides the bl_check
decorator. 
"""

import builtins
import datetime
import functools
import threading
import time
from collections import deque
from uuid import uuid4

from typeguard import typechecked

from bec_lib.bl_conditions import BeamlineCondition
from bec_lib.logger import bec_logger

logger = bec_logger.logger


class BeamlineCheckError(Exception):
    pass


class BeamlineCheckRepeat(Exception):
    pass


def bl_check(fcn):
    """Decorator to perform rpc calls."""

    @functools.wraps(fcn)
    def bl_check_wrapper(*args, **kwargs):
        client = builtins.__dict__.get("bec")
        bl_checks = client.bl_checks
        _run_with_bl_checks(bl_checks, fcn, *args, **kwargs)

    return bl_check_wrapper


def _run_with_bl_checks(bl_checks, fcn, *args, **kwargs):
    # pylint: disable=protected-access
    chk = {"id": str(uuid4()), "fcn": fcn, "args": args, "kwargs": kwargs}
    bl_checks._levels.append(chk)
    nested_call = len(bl_checks._levels) > 1
    if bl_checks._is_paused and bl_checks._beamline_checks:
        logger.warning(
            "Beamline checks are currently paused. Use `bec.bl_checks.resume()` to reactivate them."
        )
    try:
        if nested_call:
            # check if the beam was okay so far
            if not bl_checks.beam_is_okay:
                raise BeamlineCheckError("Beam is not okay.")
        else:
            bl_checks.reset()
            bl_checks.wait_for_beamline_checks()
        successful = False
        while not successful:
            try:
                successful, res = _run_on_failure(bl_checks, fcn, *args, **kwargs)

                if not bl_checks.beam_is_okay:
                    successful = False
                    bl_checks.wait_for_beamline_checks()
            except BeamlineCheckRepeat:
                successful = False
        return res

    finally:
        bl_checks._levels.pop()


def _run_on_failure(bl_checks, fcn, *args, **kwargs) -> tuple:
    try:
        res = fcn(*args, **kwargs)
        return (True, res)
    except BeamlineCheckError:
        bl_checks.wait_for_beamline_checks()
        return (False, None)


class BeamlineChecks:
    def __init__(self, client, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client
        self.send_to_scilog = True
        self._beam_is_okay = True
        self._beamline_checks = {}
        self._stop_beam_check_event = threading.Event()
        self._beam_check_thread = None
        self._started = False
        self._is_paused = False
        self._check_msgs = []
        self._levels = deque()

    @typechecked
    def register(self, check: BeamlineCondition):
        """
        Register a beamline check.

        Args:
            check (BeamlineCondition): The beamline check to register.
        """
        self._beamline_checks[check.name] = check
        setattr(self, check.name, check)

    def pause(self) -> None:
        """
        Pause beamline checks. This will disable all checks. Use `resume` to
        reactivate the checks.
        """
        self._is_paused = True

    def resume(self) -> None:
        """
        Resume all paused beamline checks.
        """
        self._is_paused = False

    def available_checks(self) -> None:
        """
        Print all available beamline checks
        """
        for name, check in self._beamline_checks.items():
            enabled = f"ENABLED: {check.enabled}"
            print(f"{name:<20} {enabled}")

    def disable_check(self, name: str) -> None:
        """
        Disable a beamline check.

        Args:
            name (str): The name of the beamline check to disable.
        """
        if name not in self._beamline_checks:
            raise ValueError(f"Beamline check {name} not registered.")
        self._beamline_checks[name].enabled = False

    def enable_check(self, name: str) -> None:
        """
        Enable a beamline check.

        Args:
            name (str): The name of the beamline check to enable.
        """
        if name not in self._beamline_checks:
            raise ValueError(f"Beamline check {name} not registered.")
        self._beamline_checks[name].enabled = True

    def disable_all_checks(self) -> None:
        """
        Disable all beamline checks.
        """
        for name in self._beamline_checks:
            self.disable_check(name)

    def enable_all_checks(self) -> None:
        """
        Enable all beamline checks.
        """
        for name in self._beamline_checks:
            self.enable_check(name)

    def _run_beamline_checks(self):
        msgs = []
        for name, check in self._beamline_checks.items():
            if not check.enabled:
                continue
            if check.run():
                continue
            msgs.append(check.on_failure_msg())
            self._beam_is_okay = False
        return msgs

    def _check_beam(self):
        while not self._stop_beam_check_event.is_set():
            self._check_msgs = self._run_beamline_checks()
            time.sleep(1)

    def start(self):
        """Start the beamline checks."""
        if self._started:
            return
        self._beam_is_okay = True

        self._beam_check_thread = threading.Thread(target=self._check_beam, daemon=True)
        self._beam_check_thread.start()
        self._started = True

    def stop(self):
        """Stop the beamline checks"""
        if not self._started:
            return

        self._stop_beam_check_event.set()
        self._beam_check_thread.join()

    def reset(self):
        self._beam_is_okay = True
        self._check_msgs = []

    @property
    def beam_is_okay(self):
        return self._beam_is_okay

    def _print_beamline_checks(self):
        for msg in self._check_msgs:
            logger.warning(msg)

    def wait_for_beamline_checks(self):
        self._print_beamline_checks()
        if self.send_to_scilog and not self.beam_is_okay:
            self._send_to_scilog(
                f"Beamline checks failed at {str(datetime.datetime.now())}: {''.join(self._check_msgs)}",
                pen="red",
            )

        self._run_beamline_checks_until_okay()

        if self.send_to_scilog:
            self._send_to_scilog(
                f"Operation resumed at {str(datetime.datetime.now())}.", pen="green"
            )

    def _run_beamline_checks_until_okay(self):
        while True:
            if self._beam_status_is_okay():
                break
            self._print_beamline_checks()
            time.sleep(5)

    def _beam_status_is_okay(self) -> bool:
        self._beam_is_okay = True
        self._check_msgs = self._run_beamline_checks()
        return self._beam_is_okay

    def _send_to_scilog(self, msg, pen="red"):
        try:
            msg = self.client.logbook.LogbookMessage()
            msg.add_text(f"<p><mark class='pen-{pen}'><strong>{msg}</strong></mark></p>").add_tag(
                ["BEC", "beam_check"]
            )
            self.client.logbook.send_logbook_message(msg)
        except Exception:
            logger.warning("Failed to send update to SciLog.")



import pytest

from bec_lib import messages
from bec_lib.scan_data import ScanData


@pytest.fixture
def scan_data():
    scan_data = ScanData()
    for ii in range(10):
        msg = messages.ScanMessage(
            point_id=ii,
            scan_id="scan_id",
            data={
                "samx": {
                    "setpoint": {"value": ii, "timestamp": ii},
                    "samx": {"value": ii, "timestamp": ii},
                }
            },
        )
        scan_data.set(ii, msg)
    yield scan_data


def test_scan_data_provides_BECMessages(scan_data):
    assert scan_data[0] == messages.ScanMessage(
        point_id=0,
        scan_id="scan_id",
        data={
            "samx": {"setpoint": {"value": 0, "timestamp": 0}, "samx": {"value": 0, "timestamp": 0}}
        },
    )
    assert scan_data[1] == messages.ScanMessage(
        point_id=1,
        scan_id="scan_id",
        data={
            "samx": {"setpoint": {"value": 1, "timestamp": 1}, "samx": {"value": 1, "timestamp": 1}}
        },
    )


def test_scan_data_signals_single_val(scan_data):
    assert scan_data.samx.setpoint.val[0] == 0
    assert scan_data.samx.setpoint.timestamps[0] == 0
    assert scan_data.samx.samx.val[0] == 0
    assert scan_data.samx.samx.timestamps[0] == 0
    assert scan_data.samx.samx.get("val")[0] == 0

    assert scan_data.samx.setpoint[0] == {"value": 0, "timestamp": 0}
    assert scan_data["samx"]["setpoint"][0] == {"value": 0, "timestamp": 0}

    assert scan_data.samx.setpoint.get(0) == {"value": 0, "timestamp": 0}
    assert scan_data["samx"].get("setpoint").get(0) == {"value": 0, "timestamp": 0}


def test_scan_data_signals_list_val(scan_data):
    assert scan_data["samx"]["setpoint"] == {ii: {"value": ii, "timestamp": ii} for ii in range(10)}
    assert scan_data.samx.setpoint == {ii: {"value": ii, "timestamp": ii} for ii in range(10)}
    assert scan_data.samx.setpoint.timestamps == [ii for ii in range(10)]

    assert scan_data["samx"]["setpoint"]["val"] == [ii for ii in range(10)]
    assert scan_data["samx"]["setpoint"]["timestamp"] == [ii for ii in range(10)]
    assert scan_data["samx"]["setpoint"].get("val") == [ii for ii in range(10)]
    assert scan_data["samx"]["setpoint"].get("timestamp") == [ii for ii in range(10)]


def test_scan_data_device_data(scan_data):
    assert scan_data["samx"] == {
        "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
    }
    assert scan_data.samx == {
        "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
    }
    assert scan_data.samx[0] == {
        "setpoint": {"value": 0, "timestamp": 0},
        "samx": {"value": 0, "timestamp": 0},
    }
    assert scan_data.get("samx") == {
        "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
    }
    assert scan_data.samx.get(0) == {
        "setpoint": {"value": 0, "timestamp": 0},
        "samx": {"value": 0, "timestamp": 0},
    }


def test_scan_data_device_data_dict_operations(scan_data):
    assert scan_data.samx.keys() == {"setpoint": 0, "samx": 0}.keys()
    assert list(scan_data.samx.values()) == [
        {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
    ]
    assert dict(scan_data.samx.items()) == dict(
        {
            "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
            "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        }.items()
    )

    assert len(scan_data.samx.setpoint) == 10


def test_scan_data_signal_dict_operations(scan_data):
    assert (
        scan_data.samx.setpoint.keys()
        == {ii: {"value": ii, "timestamp": ii} for ii in range(10)}.keys()
    )

    assert list(scan_data.samx.setpoint.values()) == list(
        {ii: {"value": ii, "timestamp": ii} for ii in range(10)}.values()
    )

    assert dict(scan_data.samx.setpoint.items()) == dict(
        {ii: {"value": ii, "timestamp": ii} for ii in range(10)}.items()
    )

    assert len(scan_data.samx.setpoint) == 10


def test_scan_data_dict_operations(scan_data):
    assert scan_data.keys() == {"samx": 0}.keys()

    assert list(scan_data.values()) == [
        {
            "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
            "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
        }
    ]
    assert dict(scan_data.items()) == dict(
        {
            "samx": {
                "setpoint": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
                "samx": {ii: {"value": ii, "timestamp": ii} for ii in range(10)},
            }
        }.items()
    )

    assert "samx" in scan_data
    assert "not_a_device" not in scan_data

    assert len(scan_data) == 10



import inspect

from bec_lib import MessageEndpoints, bec_logger, plugin_helper
from bec_lib.messages import AvailableResourceMessage
from bec_lib.signature_serializer import signature_to_dict

from . import scans as ScanServerScans

logger = bec_logger.logger


class ScanManager:

    def __init__(self, *, parent):
        """
        Scan Manager loads and manages the available scans.
        """
        self.parent = parent
        self.available_scans = {}
        self.scan_dict = {}
        self._plugins = {}
        self.load_plugins()
        self.update_available_scans()
        self.publish_available_scans()

    def load_plugins(self):
        """load scan plugins"""
        plugins = plugin_helper.get_scan_plugins()
        if not plugins:
            return
        for name, cls in plugins.items():
            if not issubclass(cls, ScanServerScans.RequestBase):
                logger.error(
                    f"Plugin {name} is not a valid scan plugin as it does not inherit from RequestBase. Skipping."
                )
                continue
            self._plugins[name] = cls
            logger.info(f"Loading scan plugin {name}")

    def update_available_scans(self):
        """load all scans and plugin scans"""
        members = inspect.getmembers(ScanServerScans)
        for member_name, cls in self._plugins.items():
            members.append((member_name, cls))

        for name, scan_cls in members:
            try:
                is_scan = issubclass(scan_cls, ScanServerScans.RequestBase)
            except TypeError:
                is_scan = False

            if not is_scan or not scan_cls.scan_name:
                logger.debug(f"Ignoring {name}")
                continue
            if scan_cls.scan_name in self.available_scans:
                logger.error(f"{scan_cls.scan_name} already exists. Skipping.")
                continue

            report_classes = [
                ScanServerScans.RequestBase,
                ScanServerScans.ScanBase,
                ScanServerScans.AsyncFlyScanBase,
                ScanServerScans.SyncFlyScanBase,
                ScanServerScans.ScanStubs,
                ScanServerScans.ScanComponent,
            ]

            for report_cls in report_classes:
                if issubclass(scan_cls, report_cls):
                    base_cls = report_cls.__name__
            self.scan_dict[scan_cls.__name__] = scan_cls
            self.available_scans[scan_cls.scan_name] = {
                "class": scan_cls.__name__,
                "base_class": base_cls,
                "arg_input": scan_cls.arg_input,
                "required_kwargs": scan_cls.required_kwargs,
                "arg_bundle_size": scan_cls.arg_bundle_size,
                "scan_report_hint": scan_cls.scan_report_hint,
                "doc": scan_cls.__doc__ or scan_cls.__init__.__doc__,
                "signature": signature_to_dict(scan_cls.__init__),
            }

    def publish_available_scans(self):
        """send all available scans to the broker"""
        self.parent.connector.set(
            MessageEndpoints.available_scans(),
            AvailableResourceMessage(resource=self.available_scans),
        )



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import py_scibec
from dotenv import dotenv_values
from py_scibec import SciBecCore

from bec_lib import MessageEndpoints, bec_logger, messages
from bec_lib.connector import ConnectorBase
from bec_server.scihub.repeated_timer import RepeatedTimer

from .config_handler import ConfigHandler
from .scibec_metadata_handler import SciBecMetadataHandler

if TYPE_CHECKING:
    from bec_server.scihub import SciHub

logger = bec_logger.logger


class SciBecConnectorError(Exception):
    pass


class SciBecConnector:
    token_expiration_time = 86400  # one day

    def __init__(self, scihub: SciHub, connector: ConnectorBase) -> None:
        self.scihub = scihub
        self.connector = connector
        self.scibec = None
        self.host = None
        self.target_bl = None
        self.ingestor = None
        self.ingestor_secret = None
        self.ro_user = None
        self.ro_user_secret = None
        self._env_configured = False
        self.scibec_info = {}
        self._config_request_handler = None
        self._metadata_handler = None
        self.config_handler = None
        self._scibec_account_thread = None
        self._start(connector)

    def _start(self, connector: ConnectorBase):
        self.connect_to_scibec()
        self.config_handler = ConfigHandler(self, connector)
        self._start_config_request_handler()
        self._start_metadata_handler()
        self._start_scibec_account_update()

    @property
    def config(self):
        """get the current service config"""
        return self.scihub.config

    def _load_environment(self):
        env_base = self.scihub.config.service_config.get("scibec", {}).get("env_file", "")
        env_file = os.path.join(env_base, ".env")
        if not os.path.exists(env_file):
            # check if there is an env file in the parent directory
            current_dir = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            env_file = os.path.join(current_dir, ".env")
            if not os.path.exists(env_file):
                return

        config = dotenv_values(env_file)
        self._update_config(**config)

    def _update_config(
        # pylint: disable=invalid-name
        self,
        SCIBEC_HOST: str = None,
        SCIBEC_TARGET: str = None,
        SCIBEC_INGESTOR: str = None,
        SCIBEC_INGESTOR_SECRET: str = None,
        SCIBEC_RO_USER: str = None,
        SCIBEC_RO_USER_SECRET: str = None,
        **kwargs,
    ) -> None:
        self.host = SCIBEC_HOST
        self.target_bl = SCIBEC_TARGET
        self.ingestor = SCIBEC_INGESTOR
        self.ingestor_secret = SCIBEC_INGESTOR_SECRET
        self.ro_user = SCIBEC_RO_USER
        self.ro_user_secret = SCIBEC_RO_USER_SECRET

        if (
            self.host
            and self.target_bl
            and self.ingestor
            and self.ingestor_secret
            and self.ro_user
            and self.ro_user_secret
        ):
            self._env_configured = True

    def _start_scibec_account_update(self) -> None:
        """
        Start a repeated timer to update the scibec account in redis
        """
        if not self._env_configured:
            return
        if not self.scibec:
            return
        try:
            self._scibec_account_update()
            self.scibec.login(username=self.ingestor, password=self.ingestor_secret)
        except Exception as exc:
            logger.warning(f"Could not connect to SciBec: {exc}")
            return
        logger.info("Starting SciBec account update timer.")
        self._scibec_account_thread = RepeatedTimer(
            self.token_expiration_time, self._scibec_account_update
        )

    def _scibec_account_update(self):
        """
        Update the scibec account in redis
        """
        logger.info("Updating SciBec account.")
        token = self.scibec.get_new_token(username=self.ro_user, password=self.ro_user_secret)
        if token:
            self.set_scibec_account(token)

    def set_scibec_account(self, token: str) -> None:
        """
        Set the scibec account in redis
        """
        self.connector.set(
            MessageEndpoints.scibec(),
            messages.CredentialsMessage(credentials={"url": self.host, "token": f"Bearer {token}"}),
        )

    def set_redis_config(self, config):
        msg = messages.AvailableResourceMessage(resource=config)
        self.connector.set(MessageEndpoints.device_config(), msg)

    def _start_metadata_handler(self) -> None:
        self._metadata_handler = SciBecMetadataHandler(self)

    def _start_config_request_handler(self) -> None:
        self._config_request_handler = self.connector.register(
            MessageEndpoints.device_config_request(),
            cb=self._device_config_request_callback,
            parent=self,
        )

    @staticmethod
    def _device_config_request_callback(msg, *, parent, **_kwargs) -> None:
        logger.info(f"Received request: {msg}")
        parent.config_handler.parse_config_request(msg.value)

    def connect_to_scibec(self):
        """
        Connect to SciBec and set the connector to the write account
        """
        self._load_environment()
        if not self._env_configured:
            logger.warning("No environment file found. Cannot connect to SciBec.")
            return
        try:
            self._update_scibec_instance()
            self._update_experiment_info()
            self._update_eaccount_in_redis()

        except Exception as exc:
            self.scibec = None
            logger.warning(f"Could not connect to SciBec: {exc}")

    def _update_scibec_instance(self):
        logger.info(f"Connecting to SciBec on {self.host}")
        self.scibec = SciBecCore(host=self.host)
        self.scibec.login(username=self.ingestor, password=self.ingestor_secret)

    def _update_experiment_info(self):
        bl_filter = py_scibec.bec.BeamlineFilterWhere(where={"name": self.target_bl})
        beamline_info = self.scibec.beamline.beamline_controller_find(bl_filter)
        if not beamline_info:
            raise SciBecConnectorError(f"Could not find a beamline with the name {self.target_bl}")
        self.scibec_info["beamline"] = beamline_info[0]
        experiment_id = beamline_info[0].active_experiment

        if not experiment_id:
            raise SciBecConnectorError(f"Could not find an active experiment on {self.target_bl}")

        experiment = self.scibec.experiment.experiment_controller_find_by_id(experiment_id)

        if not experiment:
            raise SciBecConnectorError(f"Could not find an experiment with the id {experiment_id}")
        self.scibec_info["activeExperiment"] = experiment

    def _update_eaccount_in_redis(self):
        write_account = self.scibec_info["activeExperiment"].write_account
        if write_account[0] == "p":
            write_account = write_account.replace("p", "e")
        msg = messages.VariableMessage(value=write_account)
        self.connector.set(MessageEndpoints.account(), msg)

    def shutdown(self):
        """
        Shutdown the SciBec connector
        """
        if self._scibec_account_thread:
            self._scibec_account_thread.stop()
        if self._config_request_handler:
            self._config_request_handler.shutdown()
        if self._metadata_handler:
            self._metadata_handler.shutdown()



import pytest

from bec_lib import bec_logger

### THE NEXT FIXTURE HAS TO BE RE-ACTIVATED ONCE
### OPHYD "STATUS CALLBACKS" THREADS ARE CLEANED
### (NEXT OPHYD RELEASE)
# overwrite threads_check fixture from bec_lib,
# to have it in autouse

# @pytest.fixture(autouse=True)
# def threads_check(threads_check):
#    yield
#    bec_logger.logger.remove()
###
### MEANWHILE, THIS FIXTURE WILL JUST CLEAN LOGGER
### THREADS, AND THERE WILL BE NO CHECK FOR DANGLING
### THREADS FOR DEVICE SERVER TESTS (LIKE BEFORE...)


@pytest.fixture(autouse=True)
def threads_check():
    yield
    bec_logger.logger.remove()



import io
from unittest import mock

import pytest
from rich.console import Console

from bec_ipython_client.beamline_mixin import BeamlineMixin
from bec_ipython_client.plugins.SLS.sls_info import OperatorInfo, SLSInfo


def _get_operator_messages(num: int):
    info = {f"sls_operator_messages_message{i}": {"value": f"message{i}"} for i in range(1, num)}
    info.update({f"sls_operator_date_message{i}": {"value": f"message{i}"} for i in range(1, num)})

    for i in range(num, 6):
        info.update({f"sls_operator_messages_message{i}": {"value": ""}})
        info.update({f"sls_operator_date_message{i}": {"value": ""}})

    return info


@pytest.mark.parametrize(
    "info,out",
    (
        [
            (
                _get_operator_messages(6),
                " SLS Operator messages \n┌──────────┬──────────┐\n│ Message  │ Time     │\n├──────────┼──────────┤\n│ message1 │ message1 │\n│ message2 │ message2 │\n│ message3 │ message3 │\n│ message4 │ message4 │\n│ message5 │ message5 │\n└──────────┴──────────┘\n",
            ),
            (
                _get_operator_messages(3),
                " SLS Operator messages \n┌──────────┬──────────┐\n│ Message  │ Time     │\n├──────────┼──────────┤\n│ message1 │ message1 │\n│ message2 │ message2 │\n└──────────┴──────────┘\n",
            ),
        ]
    ),
)
def test_operator_messages(info, out):
    mixin = BeamlineMixin()
    mixin._bl_info_register(OperatorInfo)
    bl_call = mixin._bl_calls[-1]
    with mock.patch.object(bl_call, "_get_operator_messages", return_value=info) as get_op_msgs:
        console = Console(file=io.StringIO(), width=120)
        with mock.patch.object(bl_call, "_get_console", return_value=console):
            mixin.bl_show_all()
            get_op_msgs.assert_called_once()
            # pylint: disable=no-member
            output = console.file.getvalue()
            assert output == out


@pytest.mark.parametrize(
    "info,out",
    (
        [
            (
                {
                    "sls_info_machine_status": {"value": "Light Available"},
                    "sls_info_injection_mode": {"value": "TOP-UP"},
                    "sls_info_current_threshold": {"value": 400.8},
                    "sls_info_current_deadband": {"value": 1.8},
                    "sls_info_filling_pattern": {"value": "Default"},
                    "sls_info_filling_life_time": {"value": 10.2},
                    "sls_info_orbit_feedback_mode": {"value": "on"},
                    "sls_info_fast_orbit_feedback": {"value": "running"},
                    "sls_info_ring_current": {"value": 401.2},
                    "sls_info_crane_usage": {"value": "OFF"},
                },
                "                 SLS Info                 \n┌──────────────────────┬─────────────────┐\n│ Key                  │ Value           │\n├──────────────────────┼─────────────────┤\n│ Machine status       │ Light Available │\n│ Injection mode       │ TOP-UP          │\n│ Ring current         │ 401.200 mA      │\n│ Current threshold    │ 400.8           │\n│ Current deadband     │ 1.8             │\n│ Filling pattern      │ Default         │\n│ SLS filling lifetime │ 10.20 h         │\n│ Orbit feedback mode  │ on              │\n│ Fast orbit feedback  │ running         │\n│ SLS crane usage      │ OFF             │\n└──────────────────────┴─────────────────┘\n",
            )
        ]
    ),
)
def test_sls_info(info, out):
    mixin = BeamlineMixin()
    mixin._bl_info_register(SLSInfo)
    bl_call = mixin._bl_calls[-1]
    with mock.patch.object(bl_call, "_get_sls_info", return_value=info) as get_sls_info:
        console = Console(file=io.StringIO(), width=120)
        with mock.patch.object(bl_call, "_get_console", return_value=console):
            mixin.bl_show_all()
            get_sls_info.assert_called_once()
            # pylint: disable=no-member
            output = console.file.getvalue()
            assert output == out



from unittest import mock

import pytest

from bec_ipython_client.callbacks.ipython_live_updates import IPythonLiveUpdates
from bec_lib import messages
from bec_lib.queue_items import QueueItem


@pytest.mark.timeout(20)
def test_live_updates_process_queue_pending(bec_client_mock):
    client = bec_client_mock
    live_updates = IPythonLiveUpdates(client)
    request_msg = messages.ScanQueueMessage(
        scan_type="grid_scan",
        parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
        queue="primary",
        metadata={"RID": "something"},
    )
    queue = QueueItem(
        scan_manager=client.queue,
        queue_id="queue_id",
        request_blocks=[request_msg],
        status="PENDING",
        active_request_block={},
        scan_id=["scan_id"],
    )
    client.queue.queue_storage.current_scan_queue = {"primary": {"status": "RUNNING"}}
    with mock.patch.object(queue, "_update_with_buffer"):
        with mock.patch(
            "bec_lib.queue_items.QueueItem.queue_position", new_callable=mock.PropertyMock
        ) as queue_pos:
            queue_pos.return_value = 2
            with mock.patch.object(
                live_updates,
                "_available_req_blocks",
                return_value=[{"report_instructions": [], "content": {"scan_type": "grid_scan"}}],
            ):
                with mock.patch.object(live_updates, "_process_report_instructions") as process:
                    with mock.patch("builtins.print") as prt:
                        res = live_updates._process_queue(queue, request_msg, "req_id")
                        prt.assert_called_once()
                        process.assert_not_called()
                    assert res is False


@pytest.mark.timeout(20)
def test_live_updates_process_queue_running(bec_client_mock):
    client = bec_client_mock
    live_updates = IPythonLiveUpdates(client)
    request_msg = messages.ScanQueueMessage(
        scan_type="grid_scan",
        parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
        queue="primary",
        metadata={"RID": "something"},
    )
    queue = QueueItem(
        scan_manager=client.queue,
        queue_id="queue_id",
        request_blocks=[request_msg],
        status="RUNNING",
        active_request_block={},
        scan_id=["scan_id"],
    )
    live_updates._active_request = request_msg
    client.queue.queue_storage.current_scan_queue = {"primary": {"status": "RUNNING"}}
    with mock.patch.object(queue, "_update_with_buffer"):
        with mock.patch(
            "bec_lib.queue_items.QueueItem.queue_position", new_callable=mock.PropertyMock
        ) as queue_pos:
            queue_pos.return_value = 2
            with mock.patch.object(
                live_updates,
                "_available_req_blocks",
                return_value=[
                    {
                        "report_instructions": [{"wait_table": 10}],
                        "content": {"scan_type": "grid_scan"},
                    }
                ],
            ):
                with mock.patch.object(live_updates, "_process_instruction") as process:
                    with mock.patch("builtins.print") as prt:
                        res = live_updates._process_queue(queue, request_msg, "req_id")
                        prt.assert_not_called()
                        process.assert_called_once_with({"wait_table": 10})
                    assert res is True


@pytest.mark.timeout(20)
def test_live_updates_process_queue_without_status(bec_client_mock):
    client = bec_client_mock
    live_updates = IPythonLiveUpdates(client)
    request_msg = messages.ScanQueueMessage(
        scan_type="grid_scan",
        parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
        queue="primary",
        metadata={"RID": "something"},
    )
    queue = QueueItem(
        scan_manager=client.queue,
        queue_id="queue_id",
        request_blocks=[request_msg],
        status=None,
        active_request_block={},
        scan_id=["scan_id"],
    )
    with mock.patch.object(queue, "_update_with_buffer"):
        assert live_updates._process_queue(queue, request_msg, "req_id") is False


@pytest.mark.timeout(20)
def test_live_updates_process_queue_without_queue_number(bec_client_mock):
    client = bec_client_mock
    live_updates = IPythonLiveUpdates(client)
    request_msg = messages.ScanQueueMessage(
        scan_type="grid_scan",
        parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
        queue="primary",
        metadata={"RID": "something"},
    )

    with mock.patch(
        "bec_lib.queue_items.QueueItem.queue_position", new_callable=mock.PropertyMock
    ) as queue_pos:
        queue = QueueItem(
            scan_manager=client.queue,
            queue_id="queue_id",
            request_blocks=[request_msg],
            status="PENDING",
            active_request_block={},
            scan_id=["scan_id"],
        )
        queue_pos.return_value = None
        with mock.patch.object(queue, "_update_with_buffer"):
            assert live_updates._process_queue(queue, request_msg, "req_id") is False


# @pytest.mark.timeout(20)
# @pytest.mark.asyncio
# def test_live_updates_process_instruction_readback(bec_client_mock):
#     client = bec_client_mock
#     live_updates = IPythonLiveUpdates(client)
#     request_msg = messages.ScanQueueMessage(
#         scan_type="grid_scan",
#         parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
#         queue="primary",
#         metadata={"RID": "something"},
#     )
#     live_updates._active_request = request_msg
#     live_updates._user_callback = []
#     client.queue.queue_storage.current_scan_queue = {"primary": {"status": "RUNNING"}}
#     with mock.patch(
#         "bec_client_mock.callbacks.ipython_live_updates.LiveUpdatesTable", new_callable=mock.Co
#     ) as table:
#         live_updates._process_instruction({"table_wait": 10})
#         table.assert_called_once_with(
#             client, report_instructions={"table_wait": 10}, request=request_msg, callbacks=[]
#         )



"""
This module contains the custom exceptions used in the BEC library.
"""


class ScanAbortion(Exception):
    """Scan abortion exception"""


class ScanInterruption(Exception):
    """Scan interruption exception"""


class ServiceConfigError(Exception):
    """Service config error"""


class DeviceConfigError(Exception):
    """Device config error"""



from unittest import mock

from bec_lib.bl_conditions import (
    FastOrbitFeedbackCondition,
    LightAvailableCondition,
    ShutterCondition,
)


def test_shutter_condition():
    device = mock.MagicMock()
    shutter_condition = ShutterCondition(device)
    shutter_condition.run()
    device.read.assert_called_once()
    assert shutter_condition.on_failure_msg() == "Check beam failed: Shutter is closed."
    assert shutter_condition.name == "shutter"


def test_light_available_condition():
    device = mock.MagicMock()
    light_available_condition = LightAvailableCondition(device)
    light_available_condition.run()
    device.read.assert_called_once()
    assert light_available_condition.on_failure_msg() == "Check beam failed: Light not available."
    assert light_available_condition.name == "light_available"


def test_fast_orbit_feedback_condition():
    device = mock.MagicMock()
    fast_orbit_feedback_condition = FastOrbitFeedbackCondition(device)
    fast_orbit_feedback_condition.run()
    device.read.assert_called_once()
    assert (
        fast_orbit_feedback_condition.on_failure_msg()
        == "Check beam failed: Fast orbit feedback is not running."
    )
    assert fast_orbit_feedback_condition.name == "fast_orbit_feedback"



"""
This module contains a class for writing pdfs.
"""

import datetime

from fpdf import FPDF


class BECPDF(FPDF):
    """Custom PDF class for BEC."""

    def header(self):
        if not hasattr(self, "title"):
            return
        # Arial bold 15
        self.set_font("Courier", "", 8)
        # Calculate width of title and position
        w = self.get_string_width(self.title) + 6
        self.set_x((210 - w) / 2)
        # Title
        self.cell(w, 9, self.title, 0, 0, "C")
        # Line break
        self.ln(10)

    def footer(self):
        # Position at 1.5 cm from bottom
        self.set_y(-15)
        # Arial italic 8
        self.set_font("Courier", "", 8)
        # Text color in gray
        self.set_text_color(128)
        # date
        self.cell(0, 10, f"BEC, {str(datetime.datetime.now())}", 0, 0, "L")
        # Page number
        self.cell(0, 10, "Page " + str(self.page_no()), 0, 0, "R")


class PDFWriter:
    """Class for writing pdfs."""

    LEFT_MARGIN = 25
    TOP_MARGIN = 20
    RIGHT_MARGIN = 20
    FONT_SIZE = 10

    def __init__(self, file: str, title: str = None, font="Courier") -> None:
        """Filer writer for pdfs.

        Args:
            file (str): full path to the output file (including file extension).
            title (str, optional): Title of the document. Defaults to None.

        """
        self.file = file
        self.title = title
        self.font = font
        self._pdf = BECPDF(orientation="P", unit="mm", format="A4")
        if self.title:
            self._pdf.set_title(self.title)
        self._pdf.add_page()
        self._pdf.set_margins(left=self.LEFT_MARGIN, top=self.TOP_MARGIN, right=self.RIGHT_MARGIN)
        self._pdf.set_font(self.font, "", self.FONT_SIZE)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def write(self, text: str):
        """add text to the pdf"""
        self._pdf.set_x(self.LEFT_MARGIN)
        self._pdf.multi_cell(0, self.FONT_SIZE // 2, text)

    def close(self):
        """close the pdf and write to disk"""
        self._pdf.output(self.file, "F")


if __name__ == "__main__":  # pragma: no cover
    header = (
        " \n" * 3
        + "  :::            :::       :::   :::   ::::    ::: ::::::::::: \n"
        + "  :+:          :+: :+:    :+:+: :+:+:  :+:+:   :+:     :+:     \n"
        + "  +:+         +:+   +:+  +:+ +:+:+ +:+ :+:+:+  +:+     +:+     \n"
        + "  +#+        +#++:++#++: +#+  +:+  +#+ +#+ +:+ +#+     +#+     \n"
        + "  +#+        +#+     +#+ +#+       +#+ +#+  +#+#+#     +#+     \n"
        + "  #+#        #+#     #+# #+#       #+# #+#   #+#+#     #+#     \n"
        + "  ########## ###     ### ###       ### ###    #### ########### \n"
    )
    with PDFWriter("test_output.pdf") as file:
        file.write(header)



# This file is the entry point for the file writer service.
# It is called either by the bec-file-writer entry point or directly from the command line.
import argparse
import os
import threading

from bec_lib import RedisConnector, ServiceConfig, bec_logger
from bec_server import file_writer

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.INFO


def main():
    """
    Launch the file writer.
    """
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--config", default="", help="path to the config file")
    clargs = parser.parse_args()
    config_path = clargs.config

    config = ServiceConfig(config_path)

    file_writer_manager = file_writer.FileWriterManager(config, RedisConnector)
    try:
        event = threading.Event()
        logger.success("Started FileWriter")
        event.wait()
    except KeyboardInterrupt:
        file_writer_manager.shutdown()


if __name__ == "__main__":
    main()



import threading


class RepeatedTimer:
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        """start the timer"""
        if not self.is_running:
            self._timer = threading.Timer(self.interval, self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        """stop the timer"""
        self._timer.cancel()
        self._timer.join()  # cancel just stops from firing, does not stop the thread or wait for it
        self.is_running = False



"""
This module provides the alarm handler class and its related functionality.
"""

from __future__ import annotations

import enum
import threading
from collections import deque
from typing import TYPE_CHECKING

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.utils import threadlocked

if TYPE_CHECKING:
    from bec_lib.redis_connector import RedisConnector


logger = bec_logger.logger


class Alarms(int, enum.Enum):
    WARNING = 0
    MINOR = 1
    MAJOR = 2


class AlarmException(Exception):
    pass


class AlarmBase(Exception):
    def __init__(
        self, alarm: messages.AlarmMessage, alarm_type: str, severity: Alarms, handled=False
    ) -> None:
        self.alarm = alarm
        self.severity = severity
        self.handled = handled
        self.alarm_type = alarm_type
        super().__init__(self.alarm.content)

    def __str__(self) -> str:
        return (
            f"An alarm has occured. Severity: {self.severity.name}. Source:"
            f" {self.alarm.content['source']}.\n{self.alarm_type}.\n\t"
            f" {self.alarm.content['msg']}"
        )


class AlarmHandler:
    def __init__(self, connector: RedisConnector) -> None:
        self.connector = connector
        self.alarms_stack = deque(maxlen=100)
        self._raised_alarms = deque(maxlen=100)
        self._lock = threading.RLock()

    def start(self):
        """start the alarm handler and its subscriptions"""
        self.connector.register(
            topics=MessageEndpoints.alarm(),
            name="AlarmHandler",
            cb=self._alarm_register_callback,
            parent=self,
        )

    @staticmethod
    def _alarm_register_callback(msg, *, parent, **_kwargs):
        parent.add_alarm(msg.value)

    @threadlocked
    def add_alarm(self, msg: messages.AlarmMessage):
        """Add a new alarm message to the stack.

        Args:
            msg (messages.AlarmMessage): Alarm message that should be added
        """
        severity = Alarms(msg.content["severity"])
        alarm = AlarmBase(
            alarm=msg, alarm_type=msg.content["alarm_type"], severity=severity, handled=False
        )
        if severity > Alarms.MINOR:
            self.alarms_stack.appendleft(alarm)
            logger.debug(alarm)
        else:
            logger.warning(alarm)

    @threadlocked
    def get_unhandled_alarms(self, severity=Alarms.WARNING) -> list:
        """Get all unhandled alarms equal or above a minimum severity.

        Args:
            severity (Alarms, optional): Minimum severity. Defaults to Alarms.WARNING.

        Returns:
            list: List of unhandled alarms

        """
        return [
            alarm for alarm in self.alarms_stack if not alarm.handled and alarm.severity >= severity
        ]

    @threadlocked
    def get_alarm(self, severity=Alarms.WARNING):
        """Get the next alarm

        Args:
            severity (Alarm, optional): Minimum severity. Defaults to Alarms.WARNING.

        Yields:
            AlarmBase: Alarm
        """
        alarms = self.get_unhandled_alarms(severity=severity)
        for alarm in alarms:
            yield alarm

    def raise_alarms(self, severity=Alarms.MAJOR):
        """Raise unhandled alarms with specified severity.

        Args:
            severity (Alarm, optional): Minimum severity. Defaults to Alarms.MAJOR.

        Raises:
            alarms: Alarm exception.
        """
        alarms = self.get_unhandled_alarms(severity=severity)
        if len(alarms) > 0:
            alarm = alarms.pop(0)
            self._raised_alarms.append(alarm)
            raise alarm

    @threadlocked
    def clear(self):
        """clear all alarms from stack"""
        self.alarms_stack.clear()

    def shutdown(self):
        """shutdown the alarm handler"""
        self.connector.shutdown()



import datetime
import threading
import time
import traceback

from bec_lib import Alarms, DeviceBase, MessageEndpoints, bec_logger, messages

from .device_validation import DeviceValidation
from .errors import DeviceMessageError, ScanAbortion
from .scan_queue import InstructionQueueItem, InstructionQueueStatus, RequestBlock

logger = bec_logger.logger


class ScanWorker(threading.Thread):
    """
    Scan worker receives device instructions and pre-processes them before sending them to the device server
    """

    def __init__(self, *, parent, queue_name: str = "primary"):
        super().__init__(daemon=True)
        self.queue_name = queue_name
        self.name = f"ScanWorker-{queue_name}"
        self.parent = parent
        self.device_manager = self.parent.device_manager
        self.connector = self.parent.connector
        self.status = InstructionQueueStatus.IDLE
        self.signal_event = threading.Event()
        self.scan_id = None
        self.scan_motors = []
        self.readout_priority = {}
        self.scan_type = None
        self.current_scan_id = None
        self.current_scan_info = None
        self._staged_devices = set()
        self.max_point_id = 0
        self._exposure_time = None
        self.current_instruction_queue_item = None
        self._last_trigger = {}
        self._groups = {}
        self.interception_msg = None
        self.reset()
        self.validate = DeviceValidation(self.device_manager.connector, self)

    def open_scan(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Open a new scan and emit a scan status message.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        if not self.scan_id:
            self.scan_id = instr.metadata.get("scan_id")
            if instr.content["parameter"].get("scan_motors") is not None:
                self.scan_motors = [
                    self.device_manager.devices[dev]
                    for dev in instr.content["parameter"].get("scan_motors")
                ]
                self.readout_priority = instr.content["parameter"].get("readout_priority", {})
            self.scan_type = instr.content["parameter"].get("scan_type")

        if not instr.metadata.get("scan_def_id"):
            self.max_point_id = 0
        num_points = self.max_point_id + instr.content["parameter"]["num_points"]
        if self.max_point_id:
            num_points += 1

        active_rb = self.current_instruction_queue_item.active_request_block

        self._initialize_scan_info(active_rb, instr, num_points)

        # only append the table_wait if the scan is not using scan_progress
        if not self.scan_report_instructions or not self.scan_report_instructions[-1].get(
            "scan_progress"
        ):
            self.scan_report_instructions.append({"table_wait": num_points})
        self.current_instruction_queue_item.parent.queue_manager.send_queue_status()

        self._send_scan_status("open")

    def close_scan(self, instr: messages.DeviceInstructionMessage, max_point_id: int) -> None:
        """
        Close a scan and emit a scan status message.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
            max_point_id (int): Maximum point ID of the scan
        """
        scan_id = instr.metadata.get("scan_id")

        if self.scan_id != scan_id:
            return

        # reset the scan ID now that the scan will be closed
        self.scan_id = None

        scan_info = self.current_scan_info
        if scan_info.get("scan_type") == "fly":
            # flyers do not increase the point_id but instead set the num_points directly
            num_points = self.current_instruction_queue_item.active_request_block.scan.num_pos
            self.current_scan_info["num_points"] = num_points

        else:
            # point_id starts at 0
            scan_info["num_points"] = max_point_id + 1

        self._send_scan_status("closed")

    def wait_for_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Wait for devices to become ready. This is a blocking call.
        Depending on the wait_type, the devices will be checked for different statuses ("idle", "read", "trigger").

        Args:
            instr (DeviceInstructionMessage): DeviceInstructionMessage
        """
        wait_type = instr.content["parameter"].get("type")

        if wait_type == "move":
            self._wait_for_idle(instr)
        elif wait_type == "read":
            self._wait_for_read(instr)
        elif wait_type == "trigger":
            self._wait_for_trigger(instr)
        else:
            logger.error("Unknown wait command")
            raise DeviceMessageError("Unknown wait command")

    def trigger_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Trigger devices by sending a trigger instruction to the device server.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        devices = [dev.name for dev in self.device_manager.devices.get_software_triggered_devices()]
        self._last_trigger = instr
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices,
                action="trigger",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )
        logger.debug(f"Triggered devices: {devices}")

    def set_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """Send device instruction to set a device to a specific value

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
        """

        # send instruction
        self.device_manager.connector.send(MessageEndpoints.device_instructions(), instr)

    def read_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Read from devices by sending a read instruction to the device server.
        This call is not blocking. Instead, a separate call to wait_for_devices is needed to wait for the devices to become ready.
        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        if instr.metadata.get("cached"):
            self._publish_readback(instr)
            return

        connector = self.device_manager.connector

        devices = instr.content.get("device")
        if devices is None:
            devices = [
                dev.name
                for dev in self.device_manager.devices.monitored_devices(
                    readout_priority=self.readout_priority
                )
            ]
        connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices,
                action="read",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )
        return

    def kickoff_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Kickoff devices by sending a kickoff instruction to the device server.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        # logger.info("kickoff")
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=instr.content.get("device"),
                action="kickoff",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )

    def complete_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Complete devices by sending a complete instruction to the device server.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        if instr.content.get("device") is None:
            devices = [dev.name for dev in self.device_manager.devices.enabled_devices]
        else:
            devices = instr.content.get("device")
        if not isinstance(devices, list):
            devices = [devices]
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices,
                action="complete",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )
        self._wait_for_status(devices, instr.metadata)

    def baseline_reading(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Perform a baseline reading by sending a read instruction to the device server.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        baseline_devices = [
            dev.name
            for dev in self.device_manager.devices.baseline_devices(
                readout_priority=self.readout_priority
            )
        ]
        params = instr.content["parameter"]
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=baseline_devices, action="read", parameter=params, metadata=instr.metadata
            ),
        )

    def pre_scan(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Perform pre-scan actions. This is a blocking call as it waits for devices to become ready again.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
        """
        devices = [dev.name for dev in self.device_manager.devices.enabled_devices]
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices,
                action="pre_scan",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )
        self._wait_for_status(devices, instr.metadata)

    def publish_data_as_read(self, instr: messages.DeviceInstructionMessage):
        """
        Publish data as read by sending a DeviceMessage to the device_read endpoint.
        This instruction replicates the behaviour of the device server when it receives a read instruction.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
        """
        connector = self.device_manager.connector
        data = instr.content["parameter"]["data"]
        devices = instr.content["device"]
        if not isinstance(devices, list):
            devices = [devices]
        if not isinstance(data, list):
            data = [data]
        for device, dev_data in zip(devices, data):
            msg = messages.DeviceMessage(signals=dev_data, metadata=instr.metadata)
            connector.set_and_publish(MessageEndpoints.device_read(device), msg)

    def send_rpc(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Send a RPC instruction to the device server.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        self.device_manager.connector.send(MessageEndpoints.device_instructions(), instr)

    def process_scan_report_instruction(self, instr):
        """
        Process a scan report instruction by appending it to the scan_report_instructions list.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        self.scan_report_instructions.append(instr.content["parameter"])
        self.current_instruction_queue_item.parent.queue_manager.send_queue_status()

    def stage_devices(self, instr: messages.DeviceInstructionMessage) -> None:
        """
        Stage devices by sending a stage instruction to the device server.
        This is a blocking call as it waits for devices to return again.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler

        """
        async_devices = self.device_manager.devices.async_devices()
        async_device_names = [dev.name for dev in async_devices]
        excluded_devices = async_devices
        excluded_devices.extend(self.device_manager.devices.on_request_devices())
        excluded_devices.extend(self.device_manager.devices.continuous_devices())
        stage_device_names_without_async = [
            dev.name
            for dev in self.device_manager.devices.enabled_devices
            if dev not in excluded_devices
        ]
        for det in async_devices:
            self.device_manager.connector.send(
                MessageEndpoints.device_instructions(),
                messages.DeviceInstructionMessage(
                    device=det.name,
                    action="stage",
                    parameter=instr.content["parameter"],
                    metadata=instr.metadata,
                ),
            )
        self._staged_devices.update(async_device_names)

        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=stage_device_names_without_async,
                action="stage",
                parameter=instr.content["parameter"],
                metadata=instr.metadata,
            ),
        )
        self._staged_devices.update(stage_device_names_without_async)
        self._wait_for_stage(staged=True, devices=async_device_names, metadata=instr.metadata)
        self._wait_for_stage(
            staged=True, devices=stage_device_names_without_async, metadata=instr.metadata
        )

    def unstage_devices(
        self, instr: messages.DeviceInstructionMessage = None, devices: list = None, cleanup=False
    ) -> None:
        """
        Unstage devices by sending a unstage instruction to the device server.
        This is a blocking call as it waits for devices to return again.

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
            devices (list): List of devices to unstage
            cleanup (bool): If True, do not wait for devices to return

        """
        if not devices:
            devices = [dev.name for dev in self.device_manager.devices.enabled_devices]
        parameter = {} if not instr else instr.content["parameter"]
        metadata = {} if not instr else instr.metadata
        self._staged_devices.difference_update(devices)
        self.device_manager.connector.send(
            MessageEndpoints.device_instructions(),
            messages.DeviceInstructionMessage(
                device=devices, action="unstage", parameter=parameter, metadata=metadata
            ),
        )
        if not cleanup:
            self._wait_for_stage(staged=False, devices=devices, metadata=metadata)

    @property
    def scan_report_instructions(self):
        """
        List of scan report instructions
        """
        req_block = self.current_instruction_queue_item.active_request_block
        return req_block.scan_report_instructions

    def _get_devices_from_instruction(
        self, instr: messages.DeviceInstructionMessage
    ) -> list[DeviceBase]:
        """Extract devices from instruction message

        Args:
            instr (DeviceInstructionMessage): DeviceInstructionMessage

        Returns:
            list[Device]: List of devices
        """
        devices = []
        if not instr.content.get("device"):
            group = instr.content["parameter"].get("group")
            if group == "primary":
                devices = self.device_manager.devices.monitored_devices(
                    readout_priority=self.readout_priority
                )
            elif group == "scan_motor":
                devices = self.scan_motors
        else:
            instr_devices = instr.content.get("device")
            if not isinstance(instr_devices, list):
                instr_devices = [instr_devices]
            devices = [self.device_manager.devices[dev] for dev in instr_devices]
        return devices

    def _add_wait_group(self, instr: messages.DeviceInstructionMessage) -> None:
        """If needed, add a wait_group. This wait_group can later be used to
        wait for instructions to complete before continuing.

        Example:
            DeviceInstructionMessage(({'device': ['samx', 'samy'], 'action': 'read', 'parameter': {'group': 'scan_motor', 'wait_group': 'scan_motor'}}, {DIID': 0,...}))

            This instruction would create a new wait_group entry for the devices samx and samy, to finish DIID 0.

        Args:
            instr (DeviceInstructionMessage): DeviceInstructionMessage

        """
        wait_group = instr.content["parameter"].get("wait_group")
        action = instr.content["action"]
        if not wait_group or action == "wait":
            return

        devices = self._get_devices_from_instruction(instr)
        DIID = instr.metadata.get("DIID")
        if DIID is None:
            raise DeviceMessageError("Device message metadata does not contain a DIID entry.")

        if wait_group in self._groups:
            self._groups[wait_group].update({dev.name: DIID for dev in devices})
        else:
            self._groups[wait_group] = {dev.name: DIID for dev in devices}

    def _check_for_failed_movements(
        self, device_status: list, devices: list, instr: messages.DeviceInstructionMessage
    ):
        if all(dev.content["success"] for dev in device_status):
            return
        ind = [dev.content["success"] for dev in device_status].index(False)
        failed_device = devices[ind]

        # make sure that this is not an old message
        matching_DIID = device_status[ind].metadata.get("DIID") >= devices[ind][1]
        matching_RID = device_status[ind].metadata.get("RID") == instr.metadata["RID"]
        if matching_DIID and matching_RID:
            last_pos_msg = self.device_manager.connector.get(
                MessageEndpoints.device_readback(failed_device[0])
            )
            last_pos = last_pos_msg.content["signals"][failed_device[0]]["value"]
            self.connector.raise_alarm(
                severity=Alarms.MAJOR,
                source=instr.content,
                msg=(
                    f"Movement of device {failed_device[0]} failed whilst trying to reach the"
                    f" target position. Last recorded position: {last_pos}"
                ),
                alarm_type="MovementFailed",
                metadata=instr.metadata,
            )
            raise ScanAbortion

    def _wait_for_idle(self, instr: messages.DeviceInstructionMessage) -> None:
        """Wait for devices to become IDLE

        Args:
            instr (DeviceInstructionMessage): Device instruction received from the scan assembler
        """
        start = datetime.datetime.now()

        wait_group = instr.content["parameter"].get("wait_group")

        if not wait_group or wait_group not in self._groups:
            return

        group_devices = [dev.name for dev in self._get_devices_from_instruction(instr)]
        wait_group_devices = [
            (dev_name, DIID)
            for dev_name, DIID in self._groups[wait_group].items()
            if dev_name in group_devices
        ]

        logger.debug(f"Waiting for devices: {wait_group}")

        while not self.validate.devices_are_ready(
            [dev for dev, _ in wait_group_devices],
            MessageEndpoints.device_req_status,
            messages.DeviceReqStatusMessage,
            instr.metadata,
            [self.validate.devices_returned_successfully, self.validate.matching_requestID],
            wait_group_devices=wait_group_devices,
            instruction=instr,
        ):
            continue

        self._groups[wait_group] = {
            dev: DIID for dev, DIID in self._groups[wait_group].items() if dev not in group_devices
        }
        logger.debug("Finished waiting")
        logger.debug(datetime.datetime.now() - start)

    def _wait_for_read(self, instr: messages.DeviceInstructionMessage) -> None:
        start = datetime.datetime.now()

        wait_group = instr.content["parameter"].get("wait_group")

        if not wait_group or wait_group not in self._groups:
            return

        group_devices = [dev.name for dev in self._get_devices_from_instruction(instr)]
        wait_group_devices = [
            (dev_name, DIID)
            for dev_name, DIID in self._groups[wait_group].items()
            if dev_name in group_devices
        ]

        logger.debug(f"Waiting for devices: {wait_group}")

        while not self.validate.devices_are_ready(
            [dev for dev, _ in wait_group_devices],
            MessageEndpoints.device_status,
            messages.DeviceStatusMessage,
            instr.metadata,
            [self.validate.devices_are_idle],
            wait_group_devices=wait_group_devices,
        ):
            continue

        self._groups[wait_group] = {
            dev: DIID for dev, DIID in self._groups[wait_group].items() if dev not in group_devices
        }
        logger.debug("Finished waiting")
        logger.debug(datetime.datetime.now() - start)

    def _wait_for_stage(self, staged: bool, devices: list, metadata: dict) -> None:
        """
        Wait for devices to become staged/unstaged

        Args:
            staged (bool): True if devices should be staged, False if they should be unstaged
            devices (list): List of devices to wait for
            metadata (dict): Metadata of the instruction
        """

        stage_validator = (
            self.validate.devices_are_staged if staged else self.validate.devices_are_unstaged
        )

        while not self.validate.devices_are_ready(
            devices,
            MessageEndpoints.device_staged,
            messages.DeviceStatusMessage,
            metadata,
            [stage_validator, self.validate.matching_requestID],
        ):
            continue

    def _wait_for_device_server(self) -> None:
        self.parent.wait_for_service("DeviceServer")

    def _wait_for_trigger(self, instr: messages.DeviceInstructionMessage) -> None:
        trigger_time = float(instr.content["parameter"]["time"]) * self.current_scan_info.get(
            "frames_per_trigger", 1
        )
        time.sleep(trigger_time)
        devices = [dev.name for dev in self.device_manager.devices.get_software_triggered_devices()]
        metadata = self._last_trigger.metadata
        self._wait_for_status(devices, metadata)

    def _wait_for_status(self, devices, metadata):
        logger_update_delay = 5
        start = time.time()
        print_status = False

        while not self.validate.devices_are_ready(
            devices,
            MessageEndpoints.device_req_status,
            messages.DeviceReqStatusMessage,
            metadata,
            [self.validate.devices_returned_successfully],
            print_status=print_status,
        ):
            if time.time() - start > logger_update_delay:
                # report the status of the devices that are not ready yet
                print_status = True
                time.sleep(1)

    def _publish_readback(
        self, instr: messages.DeviceInstructionMessage, devices: list = None
    ) -> None:
        connector = self.device_manager.connector
        if not devices:
            devices = instr.content.get("device")

        # cached readout
        readouts = self._get_readback(devices)
        pipe = connector.pipeline()
        for readout, device in zip(readouts, devices):
            msg = messages.DeviceMessage(signals=readout, metadata=instr.metadata)
            connector.set_and_publish(MessageEndpoints.device_read(device), msg, pipe)
        return pipe.execute()

    def _get_readback(self, devices: list) -> list:
        connector = self.device_manager.connector
        # cached readout
        pipe = connector.pipeline()
        for dev in devices:
            connector.get(MessageEndpoints.device_readback(dev), pipe=pipe)
        return connector.execute_pipeline(pipe)

    def _check_for_interruption(self) -> None:
        if self.status == InstructionQueueStatus.PAUSED:
            self._send_scan_status("paused")
        while self.status == InstructionQueueStatus.PAUSED:
            time.sleep(0.1)
        if self.status == InstructionQueueStatus.STOPPED:
            raise ScanAbortion

    def _initialize_scan_info(
        self, active_rb: RequestBlock, instr: messages.DeviceInstructionMessage, num_points: int
    ):
        metadata = active_rb.metadata
        self.current_scan_info = {**instr.metadata, **instr.content["parameter"]}
        self.current_scan_info.update(metadata)
        self.current_scan_info.update(
            {
                "scan_number": self.parent.scan_number,
                "dataset_number": self.parent.dataset_number,
                "exp_time": self._exposure_time,
                "frames_per_trigger": active_rb.scan.frames_per_trigger,
                "settling_time": active_rb.scan.settling_time,
                "readout_time": active_rb.scan.readout_time,
                "acquisition_config": active_rb.scan.acquisition_config,
                "scan_report_hint": active_rb.scan.scan_report_hint,
                "scan_report_devices": active_rb.scan.scan_report_devices,
                "monitor_sync": active_rb.scan.monitor_sync,
                "num_points": num_points,
            }
        )
        self.current_scan_info["scan_msgs"] = [
            str(scan_msg) for scan_msg in self.current_instruction_queue_item.scan_msgs
        ]

        self.current_scan_info["readout_priority"] = {
            "monitored": [
                dev.name
                for dev in self.device_manager.devices.monitored_devices(
                    readout_priority=self.readout_priority
                )
            ],
            "baseline": [
                dev.name
                for dev in self.device_manager.devices.baseline_devices(
                    readout_priority=self.readout_priority
                )
            ],
            "async": [
                dev.name
                for dev in self.device_manager.devices.async_devices(
                    readout_priority=self.readout_priority
                )
            ],
            "continuous": [
                dev.name
                for dev in self.device_manager.devices.continuous_devices(
                    readout_priority=self.readout_priority
                )
            ],
            "on_request": [
                dev.name
                for dev in self.device_manager.devices.on_request_devices(
                    readout_priority=self.readout_priority
                )
            ],
        }

    def _send_scan_status(self, status: str):
        current_scan_info_print = self.current_scan_info.copy()
        if current_scan_info_print.get("positions", []):
            current_scan_info_print["positions"] = "..."
        logger.info(
            f"New scan status: {self.current_scan_id} / {status} / {current_scan_info_print}"
        )
        msg = messages.ScanStatusMessage(
            scan_id=self.current_scan_id, status=status, info=self.current_scan_info
        )
        expire = None if status in ["open", "paused"] else 1800
        pipe = self.device_manager.connector.pipeline()
        self.device_manager.connector.set(
            MessageEndpoints.public_scan_info(self.current_scan_id), msg, pipe=pipe, expire=expire
        )
        self.device_manager.connector.set_and_publish(
            MessageEndpoints.scan_status(), msg, pipe=pipe
        )
        pipe.execute()

    def _process_instructions(self, queue: InstructionQueueItem) -> None:
        """
        Process scan instructions and send DeviceInstructions to OPAAS.
        For now this is an in-memory communication. In the future however,
        we might want to pass it through a dedicated Kafka topic.
        Args:
            queue: instruction queue

        Returns:

        """
        if not queue:
            return None
        self.current_instruction_queue_item = queue

        start = time.time()
        self.max_point_id = 0

        # make sure the device server is ready to receive data
        self._wait_for_device_server()

        queue.is_active = True
        try:
            for instr in queue:
                self._check_for_interruption()
                if instr is None:
                    continue
                self._exposure_time = getattr(queue.active_request_block.scan, "exp_time", None)
                self._instruction_step(instr)
        except ScanAbortion as exc:
            self._groups = {}
            if queue.stopped or not (queue.return_to_start and queue.active_request_block):
                raise ScanAbortion from exc
            queue.stopped = True
            try:
                cleanup = queue.active_request_block.scan.return_to_start()
                self.status = InstructionQueueStatus.RUNNING
                for instr in cleanup:
                    self._check_for_interruption()
                    instr.metadata["scan_id"] = queue.queue.active_rb.scan_id
                    instr.metadata["queue_id"] = queue.queue_id
                    self._instruction_step(instr)
            except Exception as exc_return_to_start:
                # if the return_to_start fails, raise the original exception
                content = traceback.format_exc()
                logger.error(content)
                self.connector.raise_alarm(
                    severity=Alarms.MAJOR,
                    source="ScanWorker",
                    msg=content,
                    alarm_type=exc_return_to_start.__class__.__name__,
                    metadata={},
                )
                raise ScanAbortion from exc
            raise ScanAbortion from exc
        except Exception as exc:
            content = traceback.format_exc()
            logger.error(content)
            self.connector.raise_alarm(
                severity=Alarms.MAJOR,
                source={"ScanWorker": "_process_instructions"},
                msg=content,
                alarm_type=exc.__class__.__name__,
                metadata={},
            )
            raise ScanAbortion from exc
        queue.is_active = False
        queue.status = InstructionQueueStatus.COMPLETED
        self.current_instruction_queue_item = None

        logger.info(f"QUEUE ITEM finished after {time.time()-start:.2f} seconds")
        self.reset()

    def _instruction_step(self, instr: messages.DeviceInstructionMessage):
        logger.debug(instr)
        action = instr.content.get("action")
        scan_def_id = instr.metadata.get("scan_def_id")
        if self.current_scan_id != instr.metadata.get("scan_id"):
            self.current_scan_id = instr.metadata.get("scan_id")

        if "point_id" in instr.metadata:
            self.max_point_id = instr.metadata["point_id"]

        self._add_wait_group(instr)

        logger.debug(f"Device instruction: {instr}")
        self._check_for_interruption()

        if action == "open_scan":
            self.open_scan(instr)
        elif action == "close_scan" and scan_def_id is None:
            self.close_scan(instr, self.max_point_id)
        elif action == "close_scan" and scan_def_id is not None:
            pass
        elif action == "open_scan_def":
            pass
        elif action == "close_scan_def":
            self.close_scan(instr, self.max_point_id)
        elif action == "wait":
            self.wait_for_devices(instr)
        elif action == "trigger":
            self.trigger_devices(instr)
        elif action == "set":
            self.set_devices(instr)
        elif action == "read":
            self.read_devices(instr)
        elif action == "kickoff":
            self.kickoff_devices(instr)
        elif action == "complete":
            self.complete_devices(instr)
        elif action == "baseline_reading":
            self.baseline_reading(instr)
        elif action == "rpc":
            self.send_rpc(instr)
        elif action == "stage":
            self.stage_devices(instr)
        elif action == "unstage":
            self.unstage_devices(instr)
        elif action == "pre_scan":
            self.pre_scan(instr)
        elif action == "publish_data_as_read":
            self.publish_data_as_read(instr)
        elif action == "scan_report_instruction":
            self.process_scan_report_instruction(instr)

        else:
            logger.warning(f"Unknown device instruction: {instr}")

    def reset(self):
        """reset the scan worker and its member variables"""
        self._groups = {}
        self.current_scan_id = ""
        self.current_scan_info = {}
        self.scan_id = None
        self.interception_msg = None
        self.scan_motors = []

    def cleanup(self):
        """perform cleanup instructions"""
        self.unstage_devices(devices=list(self._staged_devices), cleanup=True)

    def run(self):
        try:
            while not self.signal_event.is_set():
                try:
                    for queue in self.parent.queue_manager.queues[self.queue_name]:
                        self._process_instructions(queue)
                        if self.signal_event.is_set():
                            break
                        if not queue.stopped:
                            queue.append_to_queue_history()

                except ScanAbortion:
                    content = traceback.format_exc()
                    logger.error(content)
                    queue.queue.increase_scan_number()
                    if queue.return_to_start:
                        self._send_scan_status("aborted")
                    else:
                        self._send_scan_status("halted")
                    logger.info(f"Scan aborted: {queue.queue_id}")
                    queue.status = InstructionQueueStatus.STOPPED
                    queue.append_to_queue_history()
                    self.cleanup()
                    self.parent.queue_manager.queues[self.queue_name].abort()
                    self.reset()
                    self.status = InstructionQueueStatus.RUNNING

        # pylint: disable=broad-except
        except Exception as exc:
            content = traceback.format_exc()
            logger.error(content)
            self.connector.raise_alarm(
                severity=Alarms.MAJOR,
                source={"ScanWorker": "run"},
                msg=content,
                alarm_type=exc.__class__.__name__,
                metadata={},
            )
            if self.queue_name in self.parent.queue_manager.queues:
                self.parent.queue_manager.queues[self.queue_name].abort()
            self.reset()
            logger.error(f"Scan worker stopped: {exc}. Unrecoverable error.")

    def shutdown(self):
        """shutdown the scan worker"""
        self.signal_event.set()
        if self._started.is_set():
            self.join()



import csv
import os
from unittest import mock

import pytest

from bec_lib import messages
from bec_lib.scan_data import ScanData
from bec_lib.scan_items import ScanItem
from bec_lib.scan_report import ScanReport
from bec_lib.utils.rpc_utils import user_access
from bec_lib.utils.scan_utils import _extract_scan_data, _write_csv, scan_to_csv, scan_to_dict


class ScanReportMock(ScanReport):
    def __init__(self, scan_id: str) -> None:
        super().__init__()
        self.request = mock.MagicMock()
        self.request.scan.scan_id = scan_id


@pytest.fixture
def scan_data():
    scan_data = ScanData()
    for ii in range(10):
        msg = messages.ScanMessage(
            point_id=ii,
            scan_id="scan_id",
            data={
                "samx": {
                    "setpoint": {"value": ii, "timestamp": ii},
                    "samx": {"value": ii, "timestamp": ii},
                }
            },
        )
        scan_data.set(ii, msg)
    yield scan_data


@pytest.fixture
def scanitem():
    scanitem = ScanItem(
        mock.MagicMock(), queue_id="queue_id", scan_id="scan_id", scan_number=1, status="closed"
    )
    scanitem.status_message = mock.MagicMock()
    yield scanitem


class class_mock:
    USER_ACCESS = []

    @user_access
    def _func_decorated_not_in_user_access(self, *args, **kwargs):
        return None

    @user_access
    def _func_decorated_in_user_access(self, *args, **kwargs):
        return None

    def _func_not_decorated_not_in_user_access(self, *args, **kwargs):
        return None


@pytest.fixture(scope="class")
def class_factory():
    yield class_mock()


def test_user_access(class_factory):
    """Test user_access function."""
    assert class_factory.USER_ACCESS == [
        "_func_decorated_not_in_user_access",
        "_func_decorated_in_user_access",
    ]


def test__write_csv():
    """Test _write_csv function."""

    output = [["#samx", "bpm4i"], ["2.056", "100.1234"], ["0.0", "-0.12345"]]

    _write_csv(output_name="test.csv", delimiter=",", dialect=None, output=output)
    with open("test.csv", "r") as csvfile:
        csvreader = csv.reader(csvfile)
        for row, row_value in zip(csvreader, output):
            assert row == row_value

    os.remove("test.csv")


def test_scan_to_dict(scanitem, scan_data):
    """Test scan_to_dict function."""
    scanitem.data = scan_data
    return_dict = scan_to_dict(scanitem, flat=True)
    assert return_dict.keys() == {"timestamp", "value"}
    assert return_dict["value"].keys() == {"samx", "setpoint"}
    assert len(return_dict["value"]["samx"]) == 10


def test_extract_scan_data(scanitem, scan_data):
    """Test _extract_scan_data function"""
    datasetnumber = 5
    scanitem.start_time = 1620000000
    scanitem.end_time = 1620000000 + 3.5
    scanitem.status_message.return_value = messages.ScanStatusMessage(
        scan_id=scanitem.scan_id,
        status=scanitem.status,
        info={"scan_number": scanitem.scan_number, "dataset_number": datasetnumber},
    )
    scanitem.data = scan_data
    header, body = _extract_scan_data(scanitem, header=None, write_metadata=True)
    assert len(header) == 7
    assert len(body) == 10


def test_scan_to_csv():
    """Test scan_to_csv function."""
    scanreport_mock = mock.MagicMock(spec=ScanReport)
    with pytest.raises(Exception):
        scan_to_csv(
            scan_report=scanreport_mock,
            output_name=1234,
            delimiter=",",
            dialect=None,
            header=None,
            write_metadata=True,
        )
    with pytest.raises(Exception):
        scan_to_csv(
            scan_report=[scanreport_mock, scanreport_mock, scanreport_mock],
            output_name="test.csv",
            delimiter=",",
            dialect=None,
            header=None,
            write_metadata=True,
        )
    with pytest.raises(Exception):
        scan_to_csv(
            scan_report=[scanreport_mock, scanreport_mock, scanreport_mock],
            output_name="test.csv",
            delimiter=123,
            dialect=None,
            header=None,
            write_metadata=True,
        )



# pylint: skip-file
import os
from collections import defaultdict
from unittest import mock

import pydantic
import pytest
import yaml

import bec_lib
from bec_lib import messages
from bec_lib.connector import MessageObject
from bec_lib.devicemanager import DeviceConfigError

dir_path = os.path.dirname(bec_lib.__file__)


def test_device_manager_initialize(device_manager):
    with mock.patch.object(device_manager, "_get_config") as get_config:
        device_manager.initialize("")
        get_config.assert_called_once()


@pytest.mark.parametrize(
    "msg",
    [
        (messages.DeviceConfigMessage(action="update", config={"samx": {}})),
        (messages.DeviceConfigMessage(action="add", config={"samx": {}})),
        (messages.DeviceConfigMessage(action="remove", config={"samx": {}})),
    ],
)
def test_parse_config_request(device_manager, msg):
    with mock.patch.object(device_manager, "_add_device"):
        device_manager.parse_config_message(msg)


def test_config_request_update(dm_with_devices):
    device_manager = dm_with_devices
    msg = messages.DeviceConfigMessage(action="update", config={"samx": {}})
    with mock.patch.object(device_manager, "_add_device") as add_device:
        device_manager.parse_config_message(msg)

    msg = messages.DeviceConfigMessage(
        action="update", config={"samx": {"deviceConfig": {"tolerance": 1}}}
    )
    device_manager.parse_config_message(msg)
    assert device_manager.devices.samx._config["deviceConfig"]["tolerance"] == 1

    msg = messages.DeviceConfigMessage(action="update", config={"samx": {"enabled": False}})
    device_manager.parse_config_message(msg)
    assert device_manager.devices.samx._config["enabled"] is False


def test_config_request_reload(dm_with_devices):
    device_manager = dm_with_devices

    msg = messages.DeviceConfigMessage(action="reload", config=None)
    with mock.patch.object(device_manager, "_get_config") as get_config:
        device_manager.parse_config_message(msg)
        assert len(device_manager.devices) == 0
        get_config.assert_called_once()


@pytest.mark.parametrize(
    "msg,raised",
    [
        ({"action": "add", "config": {}}, True),
        ({"action": "remove", "config": {}}, True),
        ({"action": "reload", "config": {}}, False),
        ({"action": "add", "config": {"new_device": {}}}, False),
    ],
)
def test_check_request_validity(device_manager, msg, raised):
    if raised:
        with pytest.raises((DeviceConfigError, pydantic.ValidationError)):
            msg_in = messages.DeviceConfigMessage(**msg)
            device_manager.check_request_validity(msg_in)
    else:
        msg_in = messages.DeviceConfigMessage(**msg)
        device_manager.check_request_validity(msg_in)


def test_get_config_calls_load(device_manager):
    with mock.patch.object(
        device_manager, "_get_redis_device_config", return_value={"devices": [{}]}
    ) as get_redis_config:
        with mock.patch.object(device_manager, "_load_session") as load_session:
            device_manager._get_config()
            get_redis_config.assert_called_once()
            load_session.assert_called_once()


def test_get_redis_device_config(device_manager):
    with mock.patch.object(device_manager, "connector") as connector:
        connector.get.return_value = messages.AvailableResourceMessage(resource={"devices": [{}]})
        assert device_manager._get_redis_device_config() == {"devices": [{}]}


def test_get_devices_with_tags(test_config_yaml, dm_with_devices):
    config_content = test_config_yaml
    device_manager = dm_with_devices

    available_tags = defaultdict(lambda: [])
    for dev_name, dev in config_content.items():
        tags = dev.get("deviceTags")
        if tags is None:
            continue
        for tag in tags:
            available_tags[tag].append(dev_name)

    for tag, devices in available_tags.items():
        dev_list = device_manager.devices.get_devices_with_tags(tag)
        dev_names = {dev.name for dev in dev_list}
        assert dev_names == set(devices)

    assert len(device_manager.devices.get_devices_with_tags("someting")) == 0


def test_get_software_triggered_devices(test_config_yaml, dm_with_devices):
    config_content = test_config_yaml
    device_manager = dm_with_devices

    # Only eiger has softwareTrigger set to True in test config
    software_triggered_devices = []
    for dev_name, dev_cfg in config_content.items():
        if dev_cfg.get("softwareTrigger", None):
            software_triggered_devices.append(device_manager.devices.get(dev_name))

    dev_list = device_manager.devices.get_software_triggered_devices()
    assert software_triggered_devices == dev_list


def test_show_tags(test_config_yaml, dm_with_devices):
    config_content = test_config_yaml
    device_manager = dm_with_devices

    available_tags = defaultdict(lambda: [])
    for dev_name, dev in config_content.items():
        tags = dev.get("deviceTags")
        if tags is None:
            continue
        for tag in tags:
            available_tags[tag].append(dev_name)

    assert set(device_manager.devices.show_tags()) == set(available_tags.keys())


@pytest.mark.parametrize(
    "scan_motors_in,readout_priority_in",
    [([], {}), (["samx"], {}), ([], {"monitored": ["samx"]}), ([], {"baseline": ["samx"]})],
)
def test_monitored_devices_are_unique(dm_with_devices, scan_motors_in, readout_priority_in):
    device_manager = dm_with_devices
    scan_motors = [device_manager.devices.get(dev) for dev in scan_motors_in]
    devices = device_manager.devices.monitored_devices(
        scan_motors=scan_motors, readout_priority=readout_priority_in
    )
    device_names = set(dev.name for dev in devices)
    assert len(device_names) == len(devices)


@pytest.mark.parametrize(
    "scan_motors_in,readout_priority_in",
    [
        ([], {}),
        ([], {"monitored": ["samx"], "baseline": [], "on_request": []}),
        ([], {"monitored": [], "baseline": ["samx"], "on_request": []}),
        ([], {"monitored": ["samx", "samy"], "baseline": [], "on_request": ["bpm4i"]}),
    ],
)
def test_monitored_devices_with_readout_priority(
    dm_with_devices, scan_motors_in, readout_priority_in
):
    device_manager = dm_with_devices
    scan_motors = [device_manager.devices.get(dev) for dev in scan_motors_in]
    monitored_devices = device_manager.devices.monitored_devices(
        scan_motors=scan_motors, readout_priority=readout_priority_in
    )
    baseline_devices = device_manager.devices.baseline_devices(
        scan_motors=scan_motors, readout_priority=readout_priority_in
    )
    primary_device_names = set(dev.name for dev in monitored_devices)
    baseline_devices_names = set(dev.name for dev in baseline_devices)

    assert len(primary_device_names & baseline_devices_names) == 0

    assert len(set(readout_priority_in.get("on_request", [])) & baseline_devices_names) == 0
    assert len(set(readout_priority_in.get("on_request", [])) & primary_device_names) == 0


@pytest.mark.parametrize(
    "scan_motors_in,readout_priority_in",
    [
        ([], {}),
        ([], {"monitored": ["samx"], "baseline": [], "on_request": []}),
        ([], {"monitored": [], "baseline": ["samx"], "on_request": []}),
        ([], {"monitored": ["samx", "samy"], "baseline": [], "on_request": ["bpm4i"]}),
        (
            [],
            {
                "monitored": ["samx", "samy"],
                "baseline": [],
                "on_request": ["bpm4i"],
                "async": ["bpm3i"],
            },
        ),
        (
            [],
            {
                "monitored": ["samx", "samy"],
                "baseline": [],
                "on_request": ["bpm4i"],
                "async": ["bpm3i"],
                "continuous": ["bpm6i"],
            },
        ),
        (
            [],
            {
                "monitored": ["samx"],
                "baseline": ["samy"],
                "on_request": ["bpm4i"],
                "async": ["bpm3i"],
                "continuous": ["bpm6i"],
            },
        ),
    ],
)
def test_baseline_devices(dm_with_devices, scan_motors_in, readout_priority_in):
    device_manager = dm_with_devices
    scan_motors = [device_manager.devices.get(dev) for dev in scan_motors_in]
    monitored_devices = device_manager.devices.monitored_devices(
        scan_motors=scan_motors, readout_priority=readout_priority_in
    )
    baseline_devices = device_manager.devices.baseline_devices(
        scan_motors=scan_motors, readout_priority=readout_priority_in
    )
    async_devices = device_manager.devices.async_devices(readout_priority=readout_priority_in)
    continuous_devices = device_manager.devices.continuous_devices(
        readout_priority=readout_priority_in
    )
    on_request_devices = device_manager.devices.on_request_devices(
        readout_priority=readout_priority_in
    )

    primary_device_names = set(dev.name for dev in monitored_devices)
    baseline_devices_names = set(dev.name for dev in baseline_devices)
    async_devices_names = set(dev.name for dev in async_devices)
    continuous_devices_names = set(dev.name for dev in continuous_devices)
    on_request_devices_names = set(dev.name for dev in on_request_devices)

    primary_device_names.intersection(readout_priority_in.get("monitored", [])) == set(
        readout_priority_in.get("monitored", [])
    )
    baseline_devices_names.intersection(readout_priority_in.get("baseline", [])) == set(
        readout_priority_in.get("baseline", [])
    )
    async_devices_names.intersection(readout_priority_in.get("async", [])) == set(
        readout_priority_in.get("async", [])
    )
    continuous_devices_names.intersection(readout_priority_in.get("continuous", [])) == set(
        readout_priority_in.get("continuous", [])
    )
    on_request_devices_names.intersection(readout_priority_in.get("on_request", [])) == set(
        readout_priority_in.get("on_request", [])
    )

    assert len(primary_device_names & baseline_devices_names) == 0

    assert len(set(readout_priority_in.get("on_request", [])) & baseline_devices_names) == 0
    assert len(set(readout_priority_in.get("on_request", [])) & primary_device_names) == 0
    assert len(set(readout_priority_in.get("async", [])) & primary_device_names) == 0
    assert len(set(readout_priority_in.get("continuous", [])) & primary_device_names) == 0

    assert (
        set(primary_device_names).intersection(
            set(baseline_devices_names),
            set(async_devices_names),
            set(continuous_devices_names),
            set(on_request_devices_names),
        )
    ) == set()

    assert (
        set(baseline_devices_names).intersection(
            set(async_devices_names),
            set(continuous_devices_names),
            set(on_request_devices_names),
            set(primary_device_names),
        )
        == set()
    )

    assert (
        set(async_devices_names).intersection(
            set(continuous_devices_names),
            set(on_request_devices_names),
            set(primary_device_names),
            set(baseline_devices_names),
        )
        == set()
    )


def test_device_config_update_callback(dm_with_devices):
    device_manager = dm_with_devices
    dev_config_msg = messages.DeviceConfigMessage(action="update", config={"samx": {}})
    msg = MessageObject(value=dev_config_msg, topic="")

    with mock.patch.object(device_manager, "parse_config_message") as parse_config_message:
        device_manager._device_config_update_callback(msg, parent=device_manager)
        parse_config_message.assert_called_once_with(dev_config_msg)


def test_disabled_device_not_in_monitored(dm_with_devices):
    assert "motor1_disabled" in dm_with_devices.devices
    monitored_devices = dm_with_devices.devices.monitored_devices()
    assert "motor1_disabled" not in [dev.name for dev in monitored_devices]



from __future__ import annotations

import enum
import inspect
import time as ttime
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy

import numpy as np
from bec_lib import bec_logger
from lmfit import Model, models
from prettytable import PrettyTable

logger = bec_logger.logger


class SimulatedDataException(Exception):
    """Exception raised when there is an issue with the simulated data."""


class SimulationType2D(str, enum.Enum):
    """Type of simulation to steer simulated data."""

    CONSTANT = "constant"
    GAUSSIAN = "gaussian"


class NoiseType(str, enum.Enum):
    """Type of noise to add to simulated data."""

    NONE = "none"
    UNIFORM = "uniform"
    POISSON = "poisson"


class HotPixelType(str, enum.Enum):
    """Type of hot pixel to add to simulated data."""

    CONSTANT = "constant"
    FLUCTUATING = "fluctuating"


DEFAULT_PARAMS_LMFIT = {
    "c0": 1,
    "c1": 1,
    "c2": 1,
    "c3": 1,
    "c4": 1,
    "c": 100,
    "amplitude": 100,
    "center": 0,
    "sigma": 1,
}

DEFAULT_PARAMS_NOISE = {"noise": NoiseType.UNIFORM, "noise_multiplier": 10}

DEFAULT_PARAMS_MOTOR = {"ref_motor": "samx"}

DEFAULT_PARAMS_CAMERA_GAUSSIAN = {
    "amplitude": 100,
    "center_offset": np.array([0, 0]),
    "covariance": np.array([[400, 100], [100, 400]]),
}

DEFAULT_PARAMS_CAMERA_CONSTANT = {"amplitude": 100}

DEFAULT_PARAMS_HOT_PIXEL = {
    "hot_pixel_coords": np.array([[24, 24], [50, 20], [4, 40]]),
    "hot_pixel_types": [HotPixelType.FLUCTUATING, HotPixelType.CONSTANT, HotPixelType.FLUCTUATING],
    "hot_pixel_values": np.array([1e4, 1e6, 1e4]),
}


class SimulatedDataBase(ABC):
    """Abstract base class for simulated data.

    This class should be subclassed to implement the simulated data for a specific device.
    It provides the basic functionality to set and get data from the simulated data class

    ---------------------
    The class provides the following methods:

    - execute_simulation_method:    execute a method from the simulated data class or reroute execution to device proxy class
    - sim_select_model:             select the active simulation model
    - sim_params:                   get the parameters for the active simulation mdoel
    - sim_models:                   get the available simulation models
    - update_sim_state:             update the simulated state of the device
    """

    USER_ACCESS = ["sim_params", "sim_select_model", "sim_get_models", "sim_show_all"]

    def __init__(self, *args, parent=None, device_manager=None, **kwargs) -> None:
        """
        Note:
        self._model_params duplicates parameters from _params that are solely relevant for the model used.
        This facilitates easier and faster access for computing the simulated state using the lmfit package.
        """
        self.parent = parent
        self.device_manager = device_manager
        self.sim_state = defaultdict(dict)
        self.registered_proxies = getattr(self.parent, "registered_proxies", {})
        self._model = {}
        self._model_params = None
        self._params = {}

    def execute_simulation_method(self, *args, method=None, signal_name: str = "", **kwargs) -> any:
        """
        Execute either the provided method or reroutes the method execution
        to a device proxy in case it is registered in self.parentregistered_proxies.
        """
        if self.registered_proxies and self.device_manager:
            for proxy_name, signal in self.registered_proxies.items():
                if signal == signal_name or f"{self.parent.name}_{signal}" == signal_name:
                    sim_proxy = self.device_manager.devices.get(proxy_name, None)
                    if sim_proxy and sim_proxy.enabled is True:
                        method = sim_proxy.obj.lookup[self.parent.name]["method"]
                        args = sim_proxy.obj.lookup[self.parent.name]["args"]
                        kwargs = sim_proxy.obj.lookup[self.parent.name]["kwargs"]
                    break

        if method is not None:
            return method(*args, **kwargs)
        raise SimulatedDataException(f"Method {method} is not available for {self.parent.name}")

    def sim_select_model(self, model: str) -> None:
        """
        Method to select the active simulation model.
        It will initiate the model_cls and parameters for the model.

        Args:
            model (str): Name of the simulation model to select.

        """
        model_cls = self.get_model_cls(model)
        self._model = model_cls() if callable(model_cls) else model_cls
        self._params = self.get_params_for_model_cls()
        self._params.update(self._get_additional_params())
        print(self._get_table_active_simulation())

    @property
    def sim_params(self) -> dict:
        """
        Property that returns the parameters for the active simulation model. It can also
        be used to set the parameters for the active simulation updating the parameters of the model.

        Returns:
            dict: Parameters for the active simulation model.

        The following example shows how to update the noise parameter of the current simulation.
        >>> dev.<device>.sim.sim_params = {"noise": "poisson"}
        """
        return self._params

    @sim_params.setter
    def sim_params(self, params: dict):
        """
        Method to set the parameters for the active simulation model.
        """
        for k, v in params.items():
            if k in self.sim_params:
                if k == "noise":
                    self._params[k] = NoiseType(v)
                elif k == "hot_pixel_types":
                    self._params[k] = [HotPixelType(entry) for entry in v]
                else:
                    self._params[k] = v
                if isinstance(self._model, Model) and k in self._model_params:
                    self._model_params[k].value = v
            else:
                raise SimulatedDataException(f"Parameter {k} not found in {self.sim_params}.")

    def sim_get_models(self) -> list:
        """
        Method to get the all available simulation models.
        """
        return self.get_all_sim_models()

    def update_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device.

        Args:
            signal_name (str): Name of the signal to update.
            value (any): Value to update in the simulated state.
        """
        self.sim_state[signal_name]["value"] = value
        self.sim_state[signal_name]["timestamp"] = ttime.time()

    @abstractmethod
    def _get_additional_params(self) -> dict:
        """Initialize the default parameters for the noise."""

    @abstractmethod
    def get_model_cls(self, model: str) -> any:
        """
        Method to get the class for the active simulation model_cls
        """

    @abstractmethod
    def get_params_for_model_cls(self) -> dict:
        """
        Method to get the parameters for the active simulation model.
        """

    @abstractmethod
    def get_all_sim_models(self) -> list[str]:
        """
        Method to get all names from the available simulation models.

        Returns:
            list: List of available simulation models.
        """

    @abstractmethod
    def compute_sim_state(self, signal_name: str, compute_readback: bool) -> None:
        """
        Method to compute the simulated state of the device.
        """

    def _get_table_active_simulation(self, width: int = 140) -> PrettyTable:
        """Return a table with the active simulation model and parameters."""
        table = PrettyTable()
        table.title = f"Currently active model: {self._model}"
        table.field_names = ["Parameter", "Value", "Type"]
        for k, v in self.sim_params.items():
            table.add_row([k, f"{v}", f"{type(v)}"])
        table._min_width["Parameter"] = 25 if width > 75 else width // 3
        table._min_width["Type"] = 25 if width > 75 else width // 3
        table.max_table_width = width
        table._min_table_width = width

        return table

    def _get_table_method_information(self, width: int = 140) -> PrettyTable:
        """Return a table with the information about methods."""
        table = PrettyTable()
        table.max_width["Value"] = 120
        table.hrules = 1
        table.title = "Available methods within the simulation module"
        table.field_names = ["Method", "Docstring"]

        table.add_row([self.sim_get_models.__name__, f"{self.sim_get_models.__doc__}"])
        table.add_row([self.sim_select_model.__name__, self.sim_select_model.__doc__])
        table.add_row(["sim_params", self.__class__.sim_params.__doc__])
        table.max_table_width = width
        table._min_table_width = width
        table.align["Docstring"] = "l"

        return table

    def sim_show_all(self):
        """Returns a summary about the active simulation and available methods."""
        width = 150
        print(self._get_table_active_simulation(width=width))
        print(self._get_table_method_information(width=width))
        table = PrettyTable()
        table.title = "Simulation module for current device"
        table.field_names = ["All available models"]
        table.add_row([", ".join(self.get_all_sim_models())])
        table.max_table_width = width
        table._min_table_width = width
        print(table)


class SimulatedPositioner(SimulatedDataBase):
    """Simulated data class for a positioner."""

    def _init_default_additional_params(self) -> None:
        """No need to init additional parameters for Positioner."""

    def get_model_cls(self, model: str) -> any:
        """For the simulated positioners, no simulation models are currently implemented."""
        return None

    def get_params_for_model_cls(self) -> dict:
        """For the simulated positioners, no simulation models are currently implemented."""
        return {}

    def get_all_sim_models(self) -> list[str]:
        """
        For the simulated positioners, no simulation models are currently implemented.

        Returns:
            list: List of available simulation models.
        """
        return []

    def _get_additional_params(self) -> dict:
        """No need to add additional parameters for Positioner."""
        return {}

    def compute_sim_state(self, signal_name: str, compute_readback: bool) -> None:
        """
        For the simulated positioners, a computed signal is currently not used.
        The position is updated by the parent device, and readback/setpoint values
        have a jitter/tolerance introduced directly in the parent class (SimPositioner).
        """
        if compute_readback:
            method = None
            value = self.execute_simulation_method(method=method, signal_name=signal_name)
            self.update_sim_state(signal_name, value)


class SimulatedDataMonitor(SimulatedDataBase):
    """Simulated data class for a monitor."""

    def __init__(self, *args, parent=None, device_manager=None, **kwargs) -> None:
        self._model_lookup = self.init_lmfit_models()
        super().__init__(*args, parent=parent, device_manager=device_manager, **kwargs)
        self.bit_depth = self.parent.BIT_DEPTH
        self._init_default()

    def _get_additional_params(self) -> None:
        params = deepcopy(DEFAULT_PARAMS_NOISE)
        params.update(deepcopy(DEFAULT_PARAMS_MOTOR))
        return params

    def _init_default(self) -> None:
        """Initialize the default parameters for the simulated data."""
        self.sim_select_model("ConstantModel")

    def get_model_cls(self, model: str) -> any:
        """Get the class for the active simulation model."""
        if model not in self._model_lookup:
            raise SimulatedDataException(f"Model {model} not found in {self._model_lookup.keys()}.")
        return self._model_lookup[model]

    def get_all_sim_models(self) -> list[str]:
        """
        Method to get all names from the available simulation models from the lmfit.models pool.

        Returns:
            list: List of available simulation models.
        """
        return list(self._model_lookup.keys())

    def get_params_for_model_cls(self) -> dict:
        """Get the parameters for the active simulation model.

        Check if default parameters are available for lmfit parameters.

        Args:
            sim_model (str): Name of the simulation model.
        Returns:
            dict: {name: value} for the active simulation model.
        """
        rtr = {}
        params = self._model.make_params()
        for name, parameter in params.items():
            if name in DEFAULT_PARAMS_LMFIT:
                rtr[name] = DEFAULT_PARAMS_LMFIT[name]
                parameter.value = rtr[name]
            else:
                if not any([np.isnan(parameter.value), np.isinf(parameter.value)]):
                    rtr[name] = parameter.value
                else:
                    rtr[name] = 1
                    parameter.value = 1
        self._model_params = params
        return rtr

    def model_lookup(self):
        """Get available models from lmfit.models."""
        return self._model_lookup

    def init_lmfit_models(self) -> dict:
        """
        Get available models from lmfit.models.

        Exclude Gaussian2dModel, ExpressionModel, Model, SplineModel.

        Returns:
            dictionary of model name : model class pairs for available models from LMFit.
        """
        model_lookup = {}
        for name, model_cls in inspect.getmembers(models):
            try:
                is_model = issubclass(model_cls, Model)
            except TypeError:
                is_model = False
            if is_model and name not in [
                "ComplexConstantModel",
                "Gaussian2dModel",
                "ExpressionModel",
                "Model",
                "SplineModel",
            ]:
                model_lookup[name] = model_cls

        return model_lookup

    def compute_sim_state(self, signal_name: str, compute_readback: bool) -> None:
        """Update the simulated state of the device.

        It will update the value in self.sim_state with the value computed by
        the chosen simulation type.

        Args:
            signal_name (str): Name of the signal to update.
        """
        if compute_readback:
            method = self._compute
            value = self.execute_simulation_method(method=method, signal_name=signal_name)
            value = self.bit_depth(np.max(value, 0))
            self.update_sim_state(signal_name, value)

    def _compute(self, *args, **kwargs) -> int:
        """
        Compute the return value for given motor position and active model.

        Returns:
            float: Value computed by the active model.
        """
        mot_name = self.sim_params["ref_motor"]
        if self.device_manager and mot_name in self.device_manager.devices:
            motor_pos = self.device_manager.devices[mot_name].obj.read()[mot_name]["value"]
        else:
            motor_pos = 0
        method = self._model
        value = int(method.eval(params=self._model_params, x=motor_pos))
        return self._add_noise(value, self.sim_params["noise"], self.sim_params["noise_multiplier"])

    def _add_noise(self, v: int, noise: NoiseType, noise_multiplier: float) -> int:
        """
        Add the currently activated noise to the simulated data.
        If NoiseType.NONE is active, the value will be returned

        Args:
            v (int): Value to add noise to.
        Returns:
            int: Value with added noise.
        """
        if noise == NoiseType.POISSON:
            v = np.random.poisson(v)
            return v
        elif noise == NoiseType.UNIFORM:
            noise = np.ceil(np.random.uniform(0, 1) * noise_multiplier).astype(int)
            v += noise * (np.random.randint(0, 2) * 2 - 1)
            return v if v > 0 else 0
        return v


class SimulatedDataWaveform(SimulatedDataMonitor):
    """Simulated data class for a waveform.

    The class inherits from SimulatedDataMonitor,
    and overwrites the relevant methods to compute
    a simulated waveform for each point.
    """

    def _get_additional_params(self) -> None:
        params = deepcopy(DEFAULT_PARAMS_NOISE)
        return params

    def compute_sim_state(self, signal_name: str, compute_readback: bool) -> None:
        """Update the simulated state of the device.

        It will update the value in self.sim_state with the value computed by
        the chosen simulation type.

        Args:
            signal_name (str): Name of the signal to update.
        """
        if compute_readback:
            method = self._compute
            value = self.execute_simulation_method(method=method, signal_name=signal_name)
            value = self.bit_depth(value)
            self.update_sim_state(signal_name, value)

    def _compute(self, *args, **kwargs) -> np.ndarray:
        """
        Compute the return value for active model.

        Returns:
            np.array: Values computed for the activate model.
        """
        size = self.parent.waveform_shape.get()
        size = size[0] if isinstance(size, tuple) else size
        method = self._model
        value = method.eval(params=self._model_params, x=np.array(range(size)))
        value *= self.sim_params["amplitude"] / np.max(value)
        return self._add_noise(value, self.sim_params["noise"], self.sim_params["noise_multiplier"])

    def _add_noise(self, v: np.ndarray, noise: NoiseType, noise_multiplier: float) -> np.ndarray:
        """Add noise to the simulated data.

        Args:
            v (np.ndarray): Simulated data.
            noise (NoiseType): Type of noise to add.
        """
        if noise == NoiseType.POISSON:
            v = np.random.poisson(np.round(v), v.shape)
            return v
        if noise == NoiseType.UNIFORM:
            v += np.random.uniform(-noise_multiplier, noise_multiplier, v.shape)
            v[v <= 0] = 0
            return v
        if noise == NoiseType.NONE:
            return v


class SimulatedDataCamera(SimulatedDataBase):
    """Simulated class to compute data for a 2D camera."""

    def __init__(self, *args, parent=None, device_manager=None, **kwargs) -> None:
        self._model_lookup = self.init_2D_models()
        self._all_default_model_params = defaultdict(dict)
        self._init_default_camera_params()
        super().__init__(*args, parent=parent, device_manager=device_manager, **kwargs)
        self.bit_depth = self.parent.BIT_DEPTH
        self._init_default()

    def _init_default(self) -> None:
        """Initialize the default model for a simulated camera

        Use the default model "Gaussian".
        """
        self.sim_select_model(SimulationType2D.GAUSSIAN)

    def init_2D_models(self) -> dict:
        """
        Get the available models for 2D camera simulations.
        """
        model_lookup = {}
        for _, model_cls in inspect.getmembers(SimulationType2D):
            if isinstance(model_cls, SimulationType2D):
                model_lookup[model_cls.value] = model_cls
        return model_lookup

    def _get_additional_params(self) -> None:
        params = deepcopy(DEFAULT_PARAMS_NOISE)
        params.update(deepcopy(DEFAULT_PARAMS_HOT_PIXEL))
        return params

    def _init_default_camera_params(self) -> None:
        """Initiate additional params for the simulated camera."""
        self._all_default_model_params.update(
            {
                self._model_lookup[SimulationType2D.CONSTANT.value]: deepcopy(
                    DEFAULT_PARAMS_CAMERA_CONSTANT
                )
            }
        )
        self._all_default_model_params.update(
            {
                self._model_lookup[SimulationType2D.GAUSSIAN.value]: deepcopy(
                    DEFAULT_PARAMS_CAMERA_GAUSSIAN
                )
            }
        )

    def get_model_cls(self, model: str) -> any:
        """For the simulated positioners, no simulation models are currently implemented."""
        if model not in self._model_lookup:
            raise SimulatedDataException(f"Model {model} not found in {self._model_lookup.keys()}.")
        return self._model_lookup[model]

    def get_params_for_model_cls(self) -> dict:
        """For the simulated positioners, no simulation models are currently implemented."""
        return self._all_default_model_params[self._model.value]

    def get_all_sim_models(self) -> list[str]:
        """
        For the simulated positioners, no simulation models are currently implemented.

        Returns:
            list: List of available simulation models.
        """
        return [entry.value for entry in self._model_lookup.values()]

    def compute_sim_state(self, signal_name: str, compute_readback: bool) -> None:
        """Update the simulated state of the device.

        It will update the value in self.sim_state with the value computed by
        the chosen simulation type.

        Args:
            signal_name (str)       : Name of the signal to update.
            compute_readback (bool) : Flag whether to compute readback based on function hosted in SimulatedData
        """
        if compute_readback:
            if self._model == SimulationType2D.CONSTANT:
                method = "_compute_constant"
            elif self._model == SimulationType2D.GAUSSIAN:
                method = "_compute_gaussian"
            value = self.execute_simulation_method(
                signal_name=signal_name, method=getattr(self, method)
            )
        else:
            value = self._compute_empty_image()
        value = self.bit_depth(value)
        self.update_sim_state(signal_name, value)

    def _compute_empty_image(self) -> np.ndarray:
        """Computes return value for sim_type = "empty_image".

        Returns:
            float: 0
        """
        try:
            shape = self.parent.image_shape.get()
            return np.zeros(shape)
        except SimulatedDataException as exc:
            raise SimulatedDataException(
                f"Could not compute empty image for {self.parent.name} with {exc} raised. Deactivate eiger to continue."
            ) from exc

    def _compute_constant(self) -> np.ndarray:
        """Compute a return value for SimulationType2D constant."""
        try:
            shape = self.parent.image_shape.get()
            v = self.sim_params.get("amplitude") * np.ones(shape, dtype=np.float32)
            v = self._add_noise(v, self.sim_params["noise"], self.sim_params["noise_multiplier"])
            return self._add_hot_pixel(
                v,
                coords=self.sim_params["hot_pixel_coords"],
                hot_pixel_types=self.sim_params["hot_pixel_types"],
                values=self.sim_params["hot_pixel_values"],
            )
        except SimulatedDataException as exc:
            raise SimulatedDataException(
                f"Could not compute constant for {self.parent.name} with {exc} raised. Deactivate eiger to continue."
            ) from exc

    def _compute_gaussian(self) -> float:
        """Computes return value for sim_type = "gauss".

        The value is based on the parameters for the gaussian in
        self._active_params and adds noise based on the noise type.

        If computation fails, it returns 0.

        Returns: float
        """

        try:
            amp = self.sim_params.get("amplitude")
            cov = self.sim_params.get("covariance")
            cen_off = self.sim_params.get("center_offset")
            shape = self.sim_state[self.parent.image_shape.name]["value"]
            pos, offset, cov, amp = self._prepare_params_gauss(
                amp=amp, cov=cov, offset=cen_off, shape=shape
            )

            v = self._compute_multivariate_gaussian(pos=pos, cen_off=offset, cov=cov, amp=amp)
            v = self._add_noise(
                v,
                noise=self.sim_params["noise"],
                noise_multiplier=self.sim_params["noise_multiplier"],
            )
            return self._add_hot_pixel(
                v,
                coords=self.sim_params["hot_pixel_coords"],
                hot_pixel_types=self.sim_params["hot_pixel_types"],
                values=self.sim_params["hot_pixel_values"],
            )
        except SimulatedDataException as exc:
            raise SimulatedDataException(
                f"Could not compute gaussian for {self.parent.name} with {exc} raised. Deactivate eiger to continue."
            ) from exc

    def _compute_multivariate_gaussian(
        self, pos: np.ndarray | list, cen_off: np.ndarray | list, cov: np.ndarray | list, amp: float
    ) -> np.ndarray:
        """Computes and returns the multivariate Gaussian distribution.

        Args:
            pos (np.ndarray): Position of the gaussian.
            cen_off (np.ndarray): Offset from center of image for the gaussian.
            cov (np.ndarray): Covariance matrix of the gaussian.

        Returns:
            np.ndarray: Multivariate Gaussian distribution.
        """
        if isinstance(pos, list):
            pos = np.array(pos)
        if isinstance(cen_off, list):
            cen_off = np.array(cen_off)
        if isinstance(cov, list):
            cov = np.array(cov)
        dim = cen_off.shape[0]
        cov_det = np.linalg.det(cov)
        cov_inv = np.linalg.inv(cov)
        norm = np.sqrt((2 * np.pi) ** dim * cov_det)
        # This einsum call calculates (x-mu)T.Sigma-1.(x-mu) in a vectorized
        # way across all the input variables.
        fac = np.einsum("...k,kl,...l->...", pos - cen_off, cov_inv, pos - cen_off)
        v = np.exp(-fac / 2) / norm
        v *= amp / np.max(v)
        return v

    def _prepare_params_gauss(
        self, amp: float, cov: np.ndarray, offset: np.ndarray, shape: tuple
    ) -> tuple:
        """Prepare the positions for the gaussian.

        Args:
            amp (float): Amplitude of the gaussian.
            cov (np.ndarray): Covariance matrix of the gaussian.
            offset (np.ndarray): Offset from the center of the image.
            shape (tuple): Shape of the image.
        Returns:
            tuple: Positions, offset and covariance matrix for the gaussian.
        """
        x, y = np.meshgrid(
            np.linspace(-shape[0] / 2, shape[0] / 2, shape[0]),
            np.linspace(-shape[1] / 2, shape[1] / 2, shape[1]),
        )
        pos = np.empty((*x.shape, 2))
        pos[:, :, 0] = x
        pos[:, :, 1] = y

        return pos, offset, cov, amp

    def _add_noise(self, v: np.ndarray, noise: NoiseType, noise_multiplier: float) -> np.ndarray:
        """Add noise to the simulated data.

        Args:
            v (np.ndarray): Simulated data.
            noise (NoiseType): Type of noise to add.
        """
        if noise == NoiseType.POISSON:
            v = np.random.poisson(np.round(v), v.shape)
            return v
        if noise == NoiseType.UNIFORM:
            v += np.random.uniform(-noise_multiplier, noise_multiplier, v.shape)
            v[v <= 0] = 0
            return v
        if noise == NoiseType.NONE:
            return v

    def _add_hot_pixel(
        self, v: np.ndarray, coords: list, hot_pixel_types: list, values: list
    ) -> np.ndarray:
        """Add hot pixels to the simulated data.

        Args:
            v (np.ndarray): Simulated data.
            hot_pixel (dict): Hot pixel parameters.
        """
        for coord, hot_pixel_type, value in zip(coords, hot_pixel_types, values):
            if coord[0] < v.shape[0] and coord[1] < v.shape[1]:
                if hot_pixel_type == HotPixelType.CONSTANT:
                    v[coord[0], coord[1]] = value
                elif hot_pixel_type == HotPixelType.FLUCTUATING:
                    maximum = np.max(v) if np.max(v) != 0 else 1
                    if v[coord[0], coord[1]] / maximum > 0.5:
                        v[coord[0], coord[1]] = value
        return v



from bec_ipython_client import BECIPythonClient
from bec_lib import bec_logger

logger = bec_logger.logger
bec_logger.level = bec_logger.LOGLEVEL.SUCCESS

# CONFIG_PATH = "../bec_config.yaml"


# config = ServiceConfig(CONFIG_PATH)

bec = BECIPythonClient()
# bec.initialize(config, RedisConnector)
bec.start()
bec.load_high_level_interface("spec_hli")

dev = bec.device_manager.devices
scans = bec.scans

# dev.samx.readback.read()

# bec._ip.prompts.status = 1


logger.success("Started BECClient")
scans.umv(dev.samx, 5, relative=True)
# scans.line_scan(dev.samx, -5, 5, dev.samy, -5, 5, steps=10, exp_time=0.1, relative=True)

# scans.round_scan_fly(dev.flyer_sim, 0, 50, 20, 3, exp_time=0.1, relative=True)
# scans.monitor_scan(dev.samx, -100, 100, relative=False)

# from bec_ipython_client.plotting import GrumpyConnector

# bec.plotter = GrumpyConnector()
# bec.plotter.connect()


# def basic_plot(data, metadata):
#     plot_name = f"Scan {metadata['scan_number']}"
#     scan_motors = metadata.get("scan_report_devices")
#     if len(scan_motors) == 2:
#         x = data["data"][scan_motors[0]][scan_motors[0]]["value"]
#         y = data["data"][scan_motors[1]][scan_motors[1]]["value"]
#     elif len(scan_motors) == 1:
#         x = data["data"][scan_motors[0]][scan_motors[0]]["value"]
#         y = data["data"]["bpm4i"]["bpm4i"]["value"]
#     if bec.plotter.current_plot != plot_name:
#         bec.plotter.new_plot(plot_name, {})
#     bec.plotter.append_data([x, y])

# import time

# import numpy as np
# from matplotlib import pyplot as plt

# plt.ion()
# # fig = plt.figure("1")
# fig, ax = plt.subplots(figsize=(10, 8))
# line, = ax.plot([], [])
# # plt.plot([])
# x_data = []
# y_data = []

# def basic_plot(data, metadata):
#     scan_motors = metadata.get("scan_report_devices")
#     x = data["data"][scan_motors[0]][scan_motors[0]]["value"]
#     y = data["data"]["bpm4i"]["bpm4i"]["value"]
#     x_data.append(x)
#     y_data.append(y)
#     plt.plot(x_data, y_data)
#     fig.canvas.draw()
#     fig.canvas.flush_events()
#     time.sleep(0.01)

# from bec_ipython_client.plugins.XTreme.live_analysis import OTFLiveAnalysis

# cb = OTFLiveAnalysis()
# scans.line_scan(
#     dev.samy,
#     -5,
#     5,
#     steps=100,
#     exp_time=0.1,
#     relative=False,
#     callback=cb,
#     md={"polarization": "plus"},
# )
# scans.line_scan(
#     dev.samy,
#     5,
#     -5,
#     steps=100,
#     exp_time=0.1,
#     relative=False,
#     callback=cb,
#     md={"polarization": "plus"},
# )
# scans.line_scan(
#     dev.samy,
#     -5,
#     5,
#     steps=100,
#     exp_time=0.1,
#     relative=False,
#     callback=cb,
#     md={"polarization": "minus"},
# )
# scans.line_scan(
#     dev.samy,
#     5,
#     -5,
#     steps=100,
#     exp_time=0.1,
#     relative=False,
#     callback=cb,
#     md={"polarization": "minus"},
# )
# scans.line_scan(dev.samy, -5, 5, steps=100, exp_time=0.1, relative=False, callback=basic_plot)

# tomo_scan_sim()

# status = scans.fermat_scan(
#     dev.samx, -5, 5, dev.samy, -5, 5, step=1, exp_time=0.02, relative=False, hide_report=True
# )
# time.sleep(2)
# status.subscribe()

# dev.samx.low_limit = -20
# scans.round_scan_fly(dev.samx, dev.samy, 0, 50, 20, 3, exp_time=0.1, relative=True)
# def plotfunc():
#     dp = PlotAxis(bk.device_manager.connector)
#     dp.start()

# scans.umv(dev.samx, 10, relative=True)

# scans.mv(dev.samx, 20, dev.samy, -20)
# s = scans.line_scan(dev.samy, -5, 40, steps=10, exp_time=0.1)
# s = scans.round_roi_scan(dev.samx, 50, dev.samy, 20, dr=2, nth=3, exp_time=0.1)

# scan_def_id = str(uuid.uuid4())
# scans.open_interactive_scan(dev.samx, dev.samy, exp_time=0.1, md={"scan_def_id": scan_def_id})
# for ii in range(5):
#     scans.mv(dev.samx, ii, dev.samy, ii + 3, md={"scan_def_id": scan_def_id})
#     scans.interactive_scan_trigger(dev.samx, dev.samy, md={"scan_def_id": scan_def_id})
# scans.close_interactive_scan(md={"scan_def_id": scan_def_id})


# @scan_def
# def new_scan():
#     for ii in range(10):
#         scans.umv(dev.samx, ii * 10)
#         scans.fermat_scan(dev.samx, -5, 5, dev.samy, -5, 5, step=1, exp_time=0.02, relative=True)


# for ii in range(10):
#     scans.umv(dev.samx, ii * 10)
#     # scans.grid_scan(dev.samx, -5, 5, 5, dev.samy, -5, 5, 10, exp_time=0.02, relative=True)
#     scans.fermat_scan(dev.samx, -5, 5, dev.samy, -5, 5, step=1, exp_time=0.02, relative=True)

# scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02)
# s = scans.grid_scan(dev.samx, -5, 5, 100, dev.samy, -5, 5, 100, exp_time=0.0, hide_report=True)
# scans.umv(dev.samx, 0)
# scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02)


# @scans.scan_group
# def scan_with_decorator():
#     scans.umv(dev.samx, 5, relative=False)
#     scans.line_scan(dev.samx, -5, 5, steps=100, exp_time=0.1, relative=True)
#     scans.umv(dev.samx, 5, relative=False)
#     scans.line_scan(dev.samx, -5, 5, steps=100, exp_time=0.1, relative=True)
#     # scans.line_scan(dev.samx, -8, 8, steps=200, exp_time=0.1, relative=True)


# with scans.scan_def:
#     scan_with_decorator()
# with scans.dataset_id_on_hold:
#     scans.line_scan(dev.samx, -5, 5, steps=10, exp_time=0.1, relative=True)
#     scans.line_scan(dev.samx, -8, 8, steps=10, exp_time=0.1, relative=True)

# scan_def_id = str(uuid.uuid4())
# scans.open_scan_def(md={"scan_def_id": scan_def_id})
# scans.line_scan(dev.samx, -5, 5, steps=10, exp_time=0.1, md={"scan_def_id": scan_def_id})
# scans.line_scan(dev.samx, -8, 8, steps=10, exp_time=0.1, md={"scan_def_id": scan_def_id})
# scans.close_scan_def(md={"scan_def_id": scan_def_id})
# for ii in range(10):
#     scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.01)

# res = dev.samx.read()
# dev.samx.summary()
# res = dev.samx.read(cached=True, use_readback=True).get("value")
# scans.umv(dev.samx, 500)
# print(dev.samx.read(cached=True, use_readback=True))
# scans.umv(dev.samx, 1000)


# @scans.scan_group
# def alignment(*args, **kwargs):
#     scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02)
#     scans.umv(dev.samx, 10)
#     scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02)
#     scans.umv(dev.samx, 10)
#     scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02)


# with scans.scan_group:
#     scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02, relative=True)
#     scans.umv(dev.samx, 10, relative=True)

# alignment()


# scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02, md={"queue_group": queue_group})
# scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=0.02, md={"queue_group": queue_group})

# scans.grid_scan(dev.samx, -5, 5, 10, dev.samy, -5, 5, 10, exp_time=1)

# event = threading.Event()
# event.wait()
print("eos")
# p.join()
